from kratos import Generator, always_ff, posedge, negedge, reduce_add, verilog
from kratos.formal import output_btor
import tempfile
import os
import pytest
import shutil

has_yosys = shutil.which("yosys") is not None


class Parent(Generator):
    def __init__(self):
        super().__init__("Parent")
        in_ = self.input("in", 16)
        out = self.output("out", 16)
        clk = self.clock("clk")
        rst = self.reset("rst")

        class Child(Generator):
            def __init__(self):
                super().__init__("Child")
                child_in = self.input("child_in", 16)
                child_out = self.output("out", 16)
                child_clk = self.clock("clk")
                child_rst = self.reset("rst")

                @always_ff((posedge, child_clk), (negedge, child_rst))
                def code():
                    if ~child_rst:
                        child_out = 0
                    else:
                        child_out = child_in
                self.add_always(code)
        children = []
        for i in range(2):
            child = Child()
            children.append(child)
            self.add_child("child_{0}".format(i), child,
                           child_in=in_, clk=clk, rst=rst)
        self.wire(out, reduce_add(*[mod.ports.out for mod in children]))


@pytest.mark.skipif(not has_yosys, reason="yosys is not available")
def test_output_btor():
    p = Parent()
    with tempfile.TemporaryDirectory() as temp:
        btor_filename = os.path.join(temp, "test.btor")
        output_btor(p, btor_filename, quite=True)
        with open(btor_filename) as f:
            lines = f.readlines()
        assert len(lines) > 10


if __name__ == "__main__":
    test_output_btor()



import ast
import copy
import enum
import inspect
import os
import sys
import textwrap

import _kratos
import astor
from _kratos import get_frame_local

from .util import print_src


def has_format_string():
    return sys.version_info[1] >= 7


def __pretty_source(source):
    return "".join(source)


class LogicOperatorVisitor(ast.NodeTransformer):
    def __init__(self, local, global_, filename, scope_ln):
        self.local = local
        self.global_ = global_
        self.filename = filename
        self.scope_ln = scope_ln

    def get_file_line(self, ln):
        ln += self.scope_ln
        with open(self.filename) as f:
            lines = f.readlines()
        return lines[ln - 2]

    @staticmethod
    def concat_ops(func_name, values):
        # this is a recursive call
        assert len(values) > 0
        if len(values) == 1:
            return values[0]
        node = ast.Call(func=ast.Attribute(attr=func_name, value=values[0],
                                           ctx=ast.Load()),
                        args=[values[1]], keywords=[])
        values = [node] + values[2:]
        return LogicOperatorVisitor.concat_ops(func_name, values)

    def check_var(self, var):
        r = eval(astor.to_source(var), self.local, self.global_)
        if not isinstance(r, _kratos.Var):
            raise SyntaxError("Cannot mix kratos variable with normal Python values in logical ops",
                              [self.filename, self.scope_ln + var.lineno,
                               var.col_offset, self.get_file_line(var.lineno)])

    def visit_BoolOp(self, node):
        # convert the bool op into a logical function
        # we don't allow mixing bool ops with normal python values and kratos values
        if isinstance(node.op, ast.And):
            # it's an and
            values = node.values
            for i in range(len(values)):
                values[i] = self.visit(values[i])
                self.check_var(values[i])
            return self.concat_ops("and_", values)
        elif isinstance(node.op, ast.Or):
            # it's an or
            values = node.values
            for i in range(len(values)):
                values[i] = self.visit(values[i])
            # the same as and
            return self.concat_ops("or_", values)
        else:
            raise SyntaxError("Invalid logical operator", (self.filename,
                                                           self.scope_ln + node.lineno,
                                                           node.col_offset,
                                                           self.get_file_line(node.lineno)))

    def visit_UnaryOp(self, node):
        if isinstance(node.op, ast.Not):
            return ast.Call(func=ast.Attribute(attr="r_not", value=node.operand,
                                               ctx=ast.Load()),
                            args=[], keywords=[])
        return node


class StaticElaborationNodeForVisitor(ast.NodeTransformer):
    class NameVisitor(ast.NodeTransformer):
        def __init__(self, target, value):
            self.target = target
            self.value = value

        def visit_Name(self, node):
            if node.id == self.target.id:
                if isinstance(self.value, int):
                    return ast.Constant(value=self.value, lineno=node.lineno)
                elif isinstance(self.value, str):
                    return ast.Str(s=self.value, lineno=node.lineno)
                else:
                    return self.value
            return node

    class HasVar(ast.NodeVisitor):
        def __init__(self, target):
            self.has_target = False
            self.target = target
            self.legal = True

        def visit_Name(self, node):
            if node.id == self.target and self.legal:
                self.has_target = True

        def visit_JoinedStr(self, _):
            self.legal = False
            self.has_target = False

    class LoopIndexVisitor(ast.NodeVisitor):
        def __init__(self, target, scope, local_env, global_env):
            self.target = target
            self.legal = True
            self.scope = scope
            self.local_env = local_env.copy()
            # just put 0 there
            self.local_env[target.id] = 0
            self.global_env = global_env

        def visit_Subscript(self, node: ast.Index):
            if not self.legal:
                return
            s = node.slice
            has_var = StaticElaborationNodeForVisitor.HasVar(self.target.id)
            has_var.visit(s)
            if has_var.has_target:
                if not isinstance(s, ast.Index):
                    self.legal = False
                    return
                # make sure that the value is a kratos var
                value_src = astor.to_source(node.value)
                try:
                    value = eval(value_src, self.local_env, self.global_env)
                    if not isinstance(value, _kratos.Var):
                        self.legal = False
                    else:
                        # make sure we can index them
                        if value.width == 1:
                            self.legal = False
                except (AttributeError, NameError):
                    return
            elif not has_var.legal:
                self.legal = False
                return
            self.generic_visit(node)

    def __init__(self, generator, fn_src, scope, unroll_for, local, global_, filename, func_ln):
        super().__init__()
        self.generator = generator
        self.fn_src = fn_src
        self.local = local.copy()
        self.global_ = global_.copy()
        self.local["self"] = self.generator
        if "scope" in local and (not isinstance(local["scope"], Scope)):
            raise SyntaxError("scope is a reserved keyword and shall not be used in a kratos program")
        self.local["scope"] = scope
        self.target_node = {}
        self.scope = scope
        self.unroll_for = unroll_for

        self.filename = filename
        self.scope_ln = func_ln

        self.key_pair = []

    def __loop_self_var(self, target, nodes):
        valid = True
        for node in nodes:
            if valid:
                visitor = StaticElaborationNodeForVisitor.LoopIndexVisitor(target, self.scope, self.local, self.global_)
                visitor.visit(node)
                valid = visitor.legal
        return valid

    def visit_For(self, node: ast.For):
        # making sure that we don't have for/else case
        if len(node.orelse) > 0:
            # this is illegal syntax
            lines = [n.lineno for n in node.orelse]
            print_src(self.fn_src, lines)
            raise SyntaxError("Illegal Syntax: you are not allowed to use "
                              "for/else in code block")
        # get the target
        iter_ = node.iter
        iter_src = astor.to_source(iter_)
        try:
            iter_obj = eval(iter_src, self.global_, self.local)
            iter_ = list(iter_obj)
        except RuntimeError:
            print_src(self.fn_src, node.iter.lineno)
            raise SyntaxError("Unable to statically evaluate loop iter")
        for v in iter_:
            if not isinstance(v, (int, str)):
                print_src(self.fn_src, node.iter.lineno)
                raise SyntaxError("Loop iter has to be either integer or "
                                  "string, got " + str(type(v)))
        target = node.target
        if not isinstance(target, ast.Name):
            print_src(self.fn_src, node.iter.lineno)
            raise SyntaxError("Unable to parse loop "
                              "target " + astor.to_source(target))

        # if not in debug mode and we are allowed to do so
        if not self.generator.debug and not self.unroll_for and isinstance(iter_obj, range) and \
                self.__loop_self_var(target, node.body):
            # return it as a function call
            _vars = iter_obj.start, iter_obj.stop, iter_obj.step, node.lineno
            _vars = [ast.Num(n=n) for n in _vars]
            keywords = []
            if self.generator.debug:
                keywords = [ast.keyword(arg="f_ln", value=ast.Constant(value=node.lineno))]
            # we redirect the var to one of the scope vars. the index is based
            # on the line number
            index_num = node.lineno
            # create the for statement in the scope
            for_stmt = _kratos.ForStmt(target.id, iter_obj.start, iter_obj.stop, iter_obj.step)
            self.scope.for_stmt[index_num] = for_stmt
            # set the for iter var
            # redirect the variable to env
            self.scope.iter_var[index_num] = for_stmt.get_iter_var()
            index = ast.Subscript(
                slice=ast.Index(value=ast.Num(n=index_num)),
                value=ast.Attribute(
                    value=ast.Name(id="scope", ctx=ast.Load()),
                    attr="iter_var", ctx=ast.Load()))
            for_node = ast.Call(func=ast.Attribute(value=ast.Name(id="scope", ctx=ast.Load()),
                                                   attr="for_", ctx=ast.Load()),
                                args=[ast.Str(s=target.id)] + _vars,
                                keywords=keywords,
                                ctx=ast.Load())
            body_visitor = StaticElaborationNodeForVisitor.NameVisitor(target, index)
            body = []
            for i in range(len(node.body)):
                n = self.visit(body_visitor.visit(node.body[i]))
                if isinstance(n, list):
                    for nn in n:
                        body.append(nn)
                else:
                    body.append(n)
            body_node = ast.Call(func=ast.Attribute(attr="loop", value=for_node, ctx=ast.Load()),
                                 args=body, keywords=[])
            # create an entry for the target
            self.local[str(target.id)] = 0
            return self.visit(ast.Expr(body_node))
        else:
            new_node = []
            for value in iter_:
                loop_body = copy.deepcopy(node.body)
                for n in loop_body:
                    # need to replace all the reference to
                    visitor = StaticElaborationNodeForVisitor.NameVisitor(target, value)
                    n = visitor.visit(n)
                    self.key_pair.append((target.id, value))
                    n = self.visit(n)
                    self.key_pair.pop(len(self.key_pair) - 1)

                    if isinstance(n, list):
                        for n_ in n:
                            new_node.append(n_)
                            self.target_node[n_] = (target.id, value)
                    else:
                        new_node.append(n)
                        self.target_node[n] = (target.id, value)
            return new_node


class StaticElaborationNodeIfVisitor(ast.NodeTransformer):
    def __init__(self, generator, fn_src, scope, local, global_, filename, func_ln):
        super().__init__()
        self.generator = generator
        self.fn_src = fn_src
        self.local = local.copy()
        self.global_ = global_.copy()
        self.local["self"] = self.generator
        if "scope" in local and (not isinstance(local["scope"], Scope)):
            raise SyntaxError("scope is a reserved keyword and shall not be used in a kratos program")
        self.local["scope"] = scope
        self.target_node = {}
        self.scope = scope

        self.filename = filename
        self.scope_ln = func_ln

        self.key_pair = []

    def __change_if_predicate(self, node):
        if isinstance(node, ast.UnaryOp):
            # notice that if the user uses `not var`, due to Python
            # implementation, it will return True/False, we need to
            # change that into r_not call
            if isinstance(node.op, ast.Not):
                target = node.operand
                target_src = astor.to_source(target)
                target_eval = eval(target_src, self.local)
                if isinstance(target_eval, _kratos.Var):
                    return ast.Call(func=ast.Attribute(value=target, attr="r_not", ctx=ast.Load()),
                                    args=[], keywords=[], ctx=ast.Load())
            else:
                return node
        elif not isinstance(node, ast.Compare):
            return node
        op = node.ops[0]
        if not isinstance(op, ast.Eq):
            return node
        left = node.left
        left_src = astor.to_source(left)
        try:
            left_val = eval(left_src, self.local)
        except _kratos.exception.VarException:
            left_val = None

        if isinstance(left_val, _kratos.Var):
            # change it into a function all
            return ast.Call(func=ast.Attribute(value=left,
                                               attr="eq",
                                               ctx=ast.Load()),
                            args=node.comparators,
                            keywords=[],
                            ctx=ast.Load)
        return node

    def visit_If(self, node: ast.If):
        predicate = node.test
        # if it's a var comparison, we change it to eq functional call
        predicate = self.__change_if_predicate(predicate)
        # we only replace stuff if the predicate has something to do with the
        # verilog variable
        predicate_src = astor.to_source(predicate)
        has_var = False
        try:
            predicate_value = eval(predicate_src, self.global_, self.local)
        except (_kratos.exception.InvalidConversionException, _kratos.exception.UserException,
                _kratos.exception.VarException) as _:
            has_var = True
            predicate_value = None

        # if's a kratos var, we continue
        if not has_var and not isinstance(predicate_value, _kratos.Var):
            if not isinstance(predicate_value, bool):
                print_src(self.fn_src, predicate.lineno)
                raise SyntaxError("Cannot statically evaluate if predicate")
            if predicate_value:
                for i, n in enumerate(node.body):
                    if_exp = StaticElaborationNodeIfVisitor(self.generator, self.fn_src, self.scope,
                                                            self.local, self.global_, self.filename, self.scope_ln)
                    node.body[i] = if_exp.visit(n)
                return node.body
            else:
                for i, n in enumerate(node.orelse):
                    if_exp = StaticElaborationNodeIfVisitor(self.generator, self.fn_src, self.scope,
                                                            self.local, self.global_, self.filename, self.scope_ln)
                    node.orelse[i] = if_exp.visit(n)
                return node.orelse
        else:
            # need to convert the logical operators to either reduced function calls, or
            # expression or
            if_test = LogicOperatorVisitor(self.local, self.global_, self.filename, self.scope_ln)
            predicate = if_test.visit(node.test)

        expression = node.body
        else_expression = node.orelse

        if self.generator.debug:
            keywords_if = [ast.keyword(arg="f_ln",
                                       value=ast.Constant(value=node.lineno))]
            # do our best guess
            if len(else_expression) > 0:
                if else_expression[0].lineno != expression[0].lineno:
                    ln = else_expression[0].lineno + 1
                else:
                    ln = else_expression[0].lineno
                keywords_else = [ast.keyword(arg="f_ln",
                                             value=ast.Constant(value=ln))]
            else:
                keywords_else = []
        else:
            keywords_if = []
            keywords_else = []
        for key, value in self.key_pair:
            keywords_if.append(ast.keyword(arg=key, value=ast.Str(s=value)))

        if_node = ast.Call(func=ast.Attribute(value=ast.Name(id="scope",
                                                             ctx=ast.Load()),
                                              attr="if_",
                                              ctx=ast.Load()),
                           args=[predicate] + expression,
                           keywords=keywords_if,
                           ctx=ast.Load)
        else_node = ast.Call(func=ast.Attribute(attr="else_", value=if_node,
                                                ctx=ast.Load()),
                             args=else_expression, keywords=keywords_else)
        # store the mapping
        self.target_node[node] = else_node
        return self.visit(ast.Expr(value=else_node))


class AugAssignNodeVisitor(ast.NodeTransformer):
    def visit_AugAssign(self, node):
        # change any aug assign to normal assign
        return ast.Assign(targets=[node.target],
                          value=ast.BinOp(left=node.target, op=node.op,
                                          right=node.value),
                          lineno=node.lineno)


class AssignNodeVisitor(ast.NodeTransformer):
    def __init__(self, generator, debug):
        super().__init__()
        self.generator = generator
        self.debug = debug
        self.target_node = {}

    def visit_Assign(self, node):
        if len(node.targets) > 1:
            raise SyntaxError("tuple unpacking not allowed. got " +
                              astor.to_source(node))
        args = node.targets[:] + [node.value]
        if self.debug:
            args.append(ast.Constant(value=node.lineno))
        new_node = ast.Expr(
            value=ast.Call(func=ast.Attribute(
                value=ast.Name(id="scope",
                               ctx=ast.Load()),
                attr="assign",
                ctx=ast.Load()),
                args=args,
                keywords=[],
                lineno=node.lineno))
        self.target_node[node] = new_node
        return new_node


class AssertNodeVisitor(ast.NodeTransformer):
    def __init__(self, generator, debug):
        super().__init__()
        self.generator = generator
        self.debug = debug

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, ast.Name) and node.func.id == "assert_":
            # we need to transform it to scope.assert_()
            args = node.args
            if self.debug:
                args.append(ast.Constant(value=node.lineno))
            return ast.Call(func=ast.Attribute(
                value=ast.Name(id="scope", ctx=ast.Load()),
                attr="assert_",
                ctx=ast.Load()),
                args=args,
                keywords=[],
                lineno=node.lineno)
        return node


class ExceptionNodeVisitor(ast.NodeTransformer):
    def __init__(self, generator, debug):
        super().__init__()
        self.generator = generator
        self.debug = debug

    def visit_Raise(self, node: ast.Raise):
        func = node.exc
        if not isinstance(func, ast.Call):
            raise SyntaxError(astor.to_source(node) + " not supported")
        name = func.func
        if not isinstance(name, ast.Name):
            raise SyntaxError(astor.to_source(node) + " not supported")
        if name.id != "Exception":
            raise SyntaxError(astor.to_source(node) + " not supported")
        # change it to assert 0
        args = [ast.Constant(0)]
        if self.debug:
            args.append(ast.Constant(value=node.lineno))
        return ast.Call(func=ast.Attribute(value=ast.Name(id="scope", ctx=ast.Load()),
                                           attr="assert_", ctx=ast.Load()), args=args, keywords=[], lineno=node.lineno)


class ReturnNodeVisitor(ast.NodeTransformer):
    def __init__(self, scope_name, debug=False):
        self.scope_name = scope_name
        self.debug = debug

    def visit_Return(self, node: ast.Return):
        value = node.value
        args = [value]
        if self.debug:
            args.append(ast.Constant(value=node.lineno))

        return ast.Expr(value=ast.Call(func=ast.Attribute(
            value=ast.Name(id=self.scope_name,
                           ctx=ast.Load()),
            attr="return_",
            ctx=ast.Load()),
            args=args,
            keywords=[],
            lineno=node.lineno),
            lineno=node.lineno)


class GenVarLocalVisitor(ast.NodeTransformer):
    def __init__(self, key, value, scope_name):
        self.key = key
        self.value = value
        self.scope_name = scope_name

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, ast.Attribute):
            attr = node.func
            insert_keywords = False
            if attr.attr == "assign" and isinstance(attr.value, ast.Name) and attr.value.id == self.scope_name:
                insert_keywords = True
            if insert_keywords:
                keyword = ast.keyword(arg=self.key,
                                      value=ast.Str(s=self.value))
                node.keywords.append(keyword)

        return self.generic_visit(node)


def transform_always_comb_ssa(ast_tree, gen, _locals):
    class SSAScope:
        def __init__(self, var_ref):
            self.var_ref = var_ref.copy()
            self.created_vars = {}
            self.enable_cond = []

    class SSAVisitor(ast.NodeTransformer):
        # https://github.com/usagitoneko97/python-static-code-analysis/tree/master/cfg_and_ssa
        def __init__(self):
            self.vars = set()
            self.var_ref = {}

            self.phi_scope = [SSAScope({})]

        def visit_Assign(self, node):
            node.value = self.visit(node.value)
            # see if there is any target we need to be careful
            assert len(node.targets) == 1, "Unpacking not supported"
            target = node.targets[0]
            target = self.visit(target)
            node.targets = [target]
            return node

        def create_new_var(self, name):
            from .passes import Attribute
            assert name in _locals, "Only local variable scope is currently supported"
            var = _locals[name]
            new_name = gen.internal_generator.get_unique_variable_name("", name)
            new_var = gen.var_from_def(var, new_name)
            # this is used to indicate the original names
            new_var.add_attribute(Attribute.create("ssa={0}:{1}".format(name, var.name)))
            # scope is used to indicate the hierarchy of scopes so that the backend can reconstruct
            # the scopes
            new_var.add_attribute(Attribute.create("ssa-scope={0}".format(len(self.phi_scope))))
            # set the ssa enable condition
            if len(self.phi_scope) > 0:
                for phi in self.phi_scope:
                    cond = phi.enable_cond
                    for var in cond:
                        new_var.add_attribute(Attribute.create("ssa-en={0}".format(var)))
            _locals[new_name] = new_var
            self.var_ref[name] = new_name
            self.phi_scope[-1].created_vars[name] = new_name
            return new_name

        def visit_Name(self, node: ast.Name):
            if node.id not in self.vars:
                # whether it's a usage or an actual assign
                if isinstance(node.ctx, ast.Load):
                    return node
                else:
                    self.vars.add(node.id)
                    # need to create a rename
                    name = self.create_new_var(node.id)
                    node.id = name
                    return node
            else:
                if isinstance(node.ctx, ast.Store):
                    # need to rename again
                    name = self.create_new_var(node.id)
                    node.id = name
                else:
                    # use the existing name
                    assert node.id in self.var_ref
                    var_name = self.var_ref[node.id]
                    node.id = var_name
                return node

        def __find_var_def(self, name):
            for scope in reversed(self.phi_scope):
                if name in scope.created_vars:
                    return scope.created_vars[name]
            return None

        @staticmethod
        def __add_list(target, result):
            if isinstance(result, list):
                for e in result:
                    target.append(e)
            else:
                target.append(result)

        def set_enable_cond(self, node, is_else):
            node_str = astor.to_source(node)
            phi = self.phi_scope[-1]
            # add it to the enable
            if is_else:
                node_str = "~" + node_str
            node_str = node_str.strip()
            phi.enable_cond.append(node_str)

        def visit_If(self, node: ast.If):
            # push the mapping into a stack
            # need to flatten the if statement and add phi function
            test = self.visit(node.test)
            body = []
            self.phi_scope.append(SSAScope(self.var_ref))
            # compute the enable condition
            self.set_enable_cond(test, False)
            for stmt in node.body:
                stmt = self.visit(stmt)
                self.__add_list(body, stmt)
            body_scope: SSAScope = self.phi_scope[-1]
            # pop the last scope
            self.phi_scope = self.phi_scope[:-1]

            else_scope: SSAScope = SSAScope({})
            if node.orelse:
                self.phi_scope.append(SSAScope(self.var_ref))
                else_scope = self.phi_scope[-1]
                # compute the enable condition
                self.set_enable_cond(test, True)
                for stmt in node.orelse:
                    stmt = self.visit(stmt)
                    self.__add_list(body, stmt)
                # pop the last scope
                self.phi_scope = self.phi_scope[:-1]

            # if there is any variable created, need to create phi functions
            vars_ = []
            for var in body_scope.created_vars:
                if var not in vars_:
                    vars_.append(var)
            for var in else_scope.created_vars:
                if var not in vars_:
                    vars_.append(var)
            for var in vars_:
                if var in body_scope.created_vars:
                    true_var = body_scope.created_vars[var]
                    # both?
                    if var in else_scope.created_vars:
                        # create a phi node, in this, just a ternary
                        else_var = else_scope.created_vars[var]
                    else:
                        # need previous scope value
                        else_var = self.__find_var_def(var)
                        assert else_var is not None, "Latch {0} is created!".format(var)
                else:
                    else_var = else_scope.created_vars[var]
                    # need previous scope value
                    true_var = self.__find_var_def(var)
                    assert true_var is not None, "Latch {0} is created!".format(var)
                n = ast.Expr(value=ast.Call(func=ast.Name(id="ternary"), args=[test, ast.Name(id=true_var),
                                                                               ast.Name(id=else_var)],
                                            ctx=ast.Load(), keywords=[]),
                             lineno=test.lineno)
                new_name = self.create_new_var(var)
                n = ast.Assign(targets=[ast.Name(id=new_name, ctx=ast.Store())],
                               value=n, lineno=node.lineno)
                body.append(n)

            return body

    ssa = SSAVisitor()
    ast_tree = ssa.visit(ast_tree)

    # need to add all the wires for each ref
    for v1, v2 in ssa.var_ref.items():
        gen.wire(gen.vars[v1], gen.vars[v2], no_fn_ln=True)
    return ast_tree


def add_scope_context(stmt, _locals):
    for key, value in _locals.items():
        if isinstance(value, (int, float, str, bool)):
            # this is straight forward one
            stmt.add_scope_variable(key, str(value), False)
        elif isinstance(value, _kratos.Var) and len(value.name) > 0:
            # it's a var
            stmt.add_scope_variable(key, value.name, True)


class Scope:
    def __init__(self, generator, filename, ln, add_local):
        self.stmt_list = []
        self.generator = generator
        self.filename = filename
        if self.filename:
            assert os.path.isfile(self.filename)
        self.ln = ln
        self.add_local = add_local

        self._level = 0

        self.iter_var = {}
        self.for_stmt = {}

    def if_(self, target, *args, f_ln=None, **kargs):
        add_local = self.add_local

        class IfStatement:
            def __init__(self, scope):
                self._if = _kratos.IfStmt(target)
                if f_ln is not None:
                    fn_ln = (scope.filename, f_ln + scope.ln - 1)
                    target.add_fn_ln(fn_ln, True)
                    self._if.add_fn_ln(fn_ln, True)
                    self._if.then_body().add_fn_ln(fn_ln, True)
                    # this is additional info passed in
                    if add_local:
                        add_scope_context(self._if, kargs)

                self.scope = scope
                for stmt in args:
                    if hasattr(stmt, "stmt"):
                        stmt = stmt.stmt()
                    self._if.add_then_stmt(stmt)

            def else_(self, *_args, f_ln=None):
                for stmt in _args:
                    if hasattr(stmt, "stmt"):
                        self._if.add_else_stmt(stmt.stmt())
                    else:
                        self._if.add_else_stmt(stmt)
                if f_ln is not None:
                    fn_ln = (self.scope.filename, f_ln + self.scope.ln - 1)
                    self._if.else_body().add_fn_ln(fn_ln, True)
                return self

            def stmt(self):
                return self._if

            def add_scope_variable(self, name, value, is_var=False):
                self._if.add_scope_variable(name, value, is_var)

        if_stmt = IfStatement(self)
        return if_stmt

    def for_(self, var_name, start, end, step, index, f_ln=None, **kargs):
        add_local = self.add_local
        generator = self.generator

        class ForStatement:
            def __init__(self, scope):
                self.__scope = scope
                self.__for = self.__scope.for_stmt[index]
                self.__for.get_iter_var().set_generator(generator.internal_generator)
                if f_ln is not None:
                    fn_ln = (scope.filename, f_ln + scope.ln - 1)
                    self.__for.add_fn_ln(fn_ln, True)
                    self.__for.get_loop_body().add_fn_ln(fn_ln, True)
                    # this is additional info passed in
                    if add_local:
                        add_scope_context(self.__for, kargs)

            def loop(self, *args):
                for stmt in args:
                    if hasattr(stmt, "stmt"):
                        self.__for.add_stmt(stmt.stmt())
                    else:
                        self.__for.add_stmt(stmt)
                return self

            def stmt(self):
                return self.__for

        for_stmt = ForStatement(self)
        return for_stmt

    def assign(self, a, b, f_ln=None, **kargs):
        assert isinstance(a, _kratos.Var)
        try:
            stmt = a.assign(b)
        except _kratos.exception.VarException as ex:
            if f_ln is not None:
                print_src(self.filename, f_ln + self.ln - 1)
            # re-throw it
            raise ex
        if self.generator.debug:
            assert f_ln is not None
            stmt.add_fn_ln((self.filename, f_ln + self.ln - 1), True)
            if self.add_local:
                # obtain the previous call frame info
                __local = get_frame_local()
                add_scope_context(stmt, __local)
                # this is additional info passed in
                add_scope_context(stmt, kargs)
        return stmt

    def assert_(self, value, f_ln=None, **kargs):
        assert isinstance(value, (_kratos.Var, int))
        if isinstance(value, int):
            assert value == 0
            value = _kratos.constant(0, 1, False)
        stmt = _kratos.AssertValueStmt(value)
        if self.generator.debug:
            stmt.add_fn_ln((self.filename, f_ln + self.ln - 1), True)
            if self.add_local:
                # obtain the previous call frame info
                __local = get_frame_local()
                add_scope_context(stmt, __local)
                # this is additional info passed in
                add_scope_context(stmt, kargs)
        return stmt

    def add_stmt(self, stmt):
        self.stmt_list.append(stmt)

    def statements(self):
        return self.stmt_list


class FuncScope(Scope):
    def __init__(self, generator, func_name, filename, ln):
        super().__init__(generator, filename, ln, generator.debug)
        if generator is not None:
            self.__func = generator.internal_generator.function(func_name)

        self.__var_ordering = {}

    def input(self, var_name, width, is_signed=False) -> _kratos.Var:
        return self.__func.input(var_name, width, is_signed)

    def return_(self, value, f_ln=None):
        stmt = self.__func.return_stmt(value)
        if f_ln is not None:
            stmt.add_fn_ln((self.filename, f_ln + self.ln - 1), True)
        return stmt


def transform_block_comment(fn_body):
    for i in range(len(fn_body.body)):
        node = fn_body.body[i]
        if isinstance(node, ast.Expr) and isinstance(node.value, ast.Str):
            comment = node.value.s
            comment = comment.replace("\"", "").replace("'", "")
            node.value.s = comment
            fn_body.body[i] = ast.Expr(value=ast.Call(func=ast.Name(id="comment"), args=[node.value],
                                                      ctx=ast.Load(), keywords=[]))


def add_stmt_to_scope(fn_body):
    for i in range(len(fn_body.body)):
        node = fn_body.body[i]
        fn_body.body[i] = ast.Expr(
            value=ast.Call(func=ast.Attribute(value=ast.Name(id="scope", ctx=ast.Load()),
                                              attr="add_stmt", ctx=ast.Load()),
                           args=[node], keywords=[]))


def compute_target_node(for_nodes, new_nodes):
    result = {}
    for origin_node, values in for_nodes.items():
        if origin_node in new_nodes:
            target_node = new_nodes[origin_node]
            result[target_node] = values
    return result


def __ast_transform_blocks(generator, func_tree, fn_src, fn_name, scope, insert_self,
                           filename, func_ln,
                           transform_return=False, pre_locals=None, unroll_for=False,
                           apply_ssa=False):
    # pre-compute the frames
    # we have 3 frames back
    f = inspect.currentframe().f_back.f_back.f_back
    # will go one above to get the locals as well?
    if f.f_back is not None:
        _locals = f.f_back.f_locals.copy()
    else:
        _locals = {}
    _locals.update(f.f_locals)
    _globals = f.f_globals.copy()

    if pre_locals is not None:
        _locals.update(pre_locals)

    debug = generator.debug
    fn_body = func_tree.body[0]

    func_args = fn_body.args.args
    # add out scope to the arg list to capture all the statements
    func_args.append(ast.arg(arg="scope", annotation=None))

    if transform_return:
        return_visitor = ReturnNodeVisitor("scope", generator.debug)
        return_visitor.visit(fn_body)

    # change to block comment to comment_
    transform_block_comment(fn_body)

    # transform aug assign
    aug_assign_visitor = AugAssignNodeVisitor()
    fn_body = aug_assign_visitor.visit(fn_body)

    # if there is a need to apply ssa
    if apply_ssa:
        # only elaborate for
        for_visitor = StaticElaborationNodeForVisitor(generator, fn_src, scope, True, _locals,
                                                      _globals, filename, func_ln)
        fn_body = for_visitor.visit(fn_body)
        target_nodes = for_visitor.target_node
        fn_body = transform_always_comb_ssa(fn_body, generator, _locals)
    else:
        # static eval for loop and if statement
        for_visitor = StaticElaborationNodeForVisitor(generator, fn_src, scope, unroll_for, _locals,
                                                      _globals, filename, func_ln)
        fn_body = for_visitor.visit(fn_body)
        if_visitor = StaticElaborationNodeIfVisitor(generator, fn_src, scope, _locals, _globals, filename, func_ln)
        fn_body = if_visitor.visit(fn_body)
        target_nodes = compute_target_node(for_visitor.target_node, if_visitor.target_node)

    # transform assign
    assign_visitor = AssignNodeVisitor(generator, debug)
    fn_body = assign_visitor.visit(fn_body)
    ast.fix_missing_locations(fn_body)

    if apply_ssa:
        target_nodes = compute_target_node(target_nodes, assign_visitor.target_node)

    # transform the assert_ function to get fn_ln
    assert_visitor = AssertNodeVisitor(generator, debug)
    fn_body = assert_visitor.visit(fn_body)
    exception_visitor = ExceptionNodeVisitor(generator, debug)
    fn_body = exception_visitor.visit(fn_body)

    # mark the local variables
    for node, (key, value) in target_nodes.items():
        assign_local_visitor = GenVarLocalVisitor(key, value, "scope")
        assign_local_visitor.visit(node)

    # add stmt to the scope
    add_stmt_to_scope(fn_body)

    # add code to run it
    if insert_self:
        args = [ast.Name(id="_self", ctx=ast.Load())]
    else:
        args = []
    args.append(ast.Name(id="_scope", ctx=ast.Load()))
    call_node = ast.Call(func=ast.Name(id=fn_name, ctx=ast.Load()),
                         args=args, keywords=[], ctx=ast.Load)
    func_tree.body.append(ast.Expr(value=call_node))
    return _locals, _globals


def inject_import_code(code_src):
    line1 = "import kratos"
    line2 = "from kratos import *"
    return "\n".join([line1, line2, code_src])


class CodeBlockType(enum.Enum):
    Sequential = enum.auto()
    Combinational = enum.auto()
    Initial = enum.auto()
    Latch = enum.auto()


class AlwaysWrapper:
    def __init__(self, fn):
        self.fn = fn
        assert callable(fn)
        args = extract_arg_name_order_from_fn(fn)
        self.args = set()
        for name in args.values():
            if name != "self":
                self.args.add(name)

    def __call__(self, *args, **kwargs):
        raise SyntaxError("always block cannot be called normally. "
                          "Please do self.add_block()")


def filter_fn_args(fn_args):
    args = []
    for arg in fn_args:
        assert isinstance(arg, ast.arg), astor.to_source(arg) + " is not an argument"
        if arg.arg == "self":
            args.append(arg)
    return args


def transform_stmt_block(generator, fn, unroll_for=False, apply_ssa=False, fn_ln=None, kargs=None):
    if kargs is None:
        kargs = dict()
    env_kargs = dict()

    if callable(fn) or isinstance(fn, AlwaysWrapper):
        if isinstance(fn, AlwaysWrapper):
            assert len(kargs) >= len(fn.args), "missing arguments in function"
            for name in fn.args:
                assert name in kargs, name + " not found in function args"
                env_kargs[name] = kargs[name]
            fn = fn.fn
        else:
            import warnings
            warnings.warn("Bare function code as always block will be "
                          "deprecated soon. Please use @always_ff or "
                          "@always_comb", SyntaxWarning)
            print_src(get_fn(fn), get_ln(fn))
        fn_src = inspect.getsource(fn)
        fn_name = fn.__name__
        func_tree = ast.parse(textwrap.dedent(fn_src))
    else:
        assert isinstance(fn, ast.FunctionDef)
        # user directly passed in ast nodes
        assert fn_ln is not None
        fn_name = fn.name
        func_tree = ast.Module(body=[fn])
        fn_src = astor.to_source(fn)

    fn_body = func_tree.body[0]
    # needs debug
    debug = generator.debug
    store_local = debug and fn_ln is None
    filename = get_fn(fn) if fn_ln is None else fn_ln[0]
    ln = get_ln(fn) if fn_ln is None else fn_ln[1]

    # extract the sensitivity list from the decorator
    blk_type, sensitivity = extract_sensitivity_from_dec(fn_body.decorator_list, fn_name)
    # remove the decorator
    fn_body.decorator_list = []
    # check the function args. it should only has one self now
    func_args = filter_fn_args(fn_body.args.args)
    insert_self = len(func_args) == 1
    fn_body.args.args = func_args
    # creating the scope here
    scope = Scope(generator, filename, ln, store_local)

    apply_ssa = blk_type == CodeBlockType.Combinational and apply_ssa

    _locals, _globals = __ast_transform_blocks(generator, func_tree, fn_src,
                                               fn_name, scope,
                                               insert_self, filename, ln,
                                               unroll_for=unroll_for,
                                               apply_ssa=apply_ssa,
                                               pre_locals=env_kargs)

    src = astor.to_source(func_tree, pretty_source=__pretty_source)
    src = inject_import_code(src)
    code_obj = compile(src, "<ast>", "exec")

    # notice that this ln is an offset
    _locals.update({"_self": generator, "_scope": scope})
    # use user specified args
    _locals.update(env_kargs)
    for name in env_kargs:
        kargs.pop(name)
    _globals.update(_locals)
    exec(code_obj, _globals)
    stmts = scope.statements()
    return blk_type, sensitivity, stmts


def transform_function_block(generator, fn, arg_types):
    fn_src = inspect.getsource(fn)
    fn_name = fn.__name__
    func_tree = ast.parse(textwrap.dedent(fn_src))
    fn_body = func_tree.body[0]
    # needs debug
    debug = generator.debug

    # remove the decorator
    fn_body.decorator_list = []

    # check the function args. it should only has one self now
    func_args = fn_body.args.args
    insert_self = func_args[0].arg == "self"
    # only keep self
    fn_body.args.args = [func_args[0]]
    # add function args now
    filename = get_fn(fn)
    ln = get_ln(fn)
    scope = FuncScope(generator, fn_name, filename, ln)
    # add var creations
    arg_order = extract_arg_name_order_from_ast(func_args)
    var_body = declare_var_definition(arg_types, arg_order)
    var_src = astor.to_source(ast.Module(body=var_body))
    pre_locals = {"_scope": scope}
    var_code_obj = compile(var_src, "<ast>", "exec")
    exec(var_code_obj, pre_locals)
    _locals, _globals = __ast_transform_blocks(generator, func_tree, fn_src,
                                               fn_name, scope, insert_self,
                                               filename, ln,
                                               transform_return=True,
                                               pre_locals=pre_locals)

    src = astor.to_source(func_tree)
    src = inject_import_code(src)
    code_obj = compile(src, "<ast>", "exec")

    _locals.update({"_self": generator, "_scope": scope})
    _globals.update(_locals)
    exec(code_obj, _globals)
    stmts = scope.statements()
    return arg_order, stmts


def declare_var_definition(var_def, arg_order):
    body = []
    for idx, name in arg_order.items():
        width, is_signed = var_def[idx]
        body.append(ast.Assign(targets=[ast.Name(id=name)],
                               value=ast.Call(func=ast.Attribute(value=ast.Name(id="_scope", ctx=ast.Load()),
                                                                 attr="input", ctx=ast.Load()),
                                              args=[ast.Str(s=name), ast.Constant(value=width),
                                                    ast.NameConstant(value=is_signed)],
                                              keywords=[])))
    return body


def extract_arg_name_order_from_ast(func_args):
    result = {}
    for idx, arg in enumerate(func_args):
        if arg.arg != "self":
            result[len(result)] = arg.arg
    return result


def extract_arg_name_order_from_fn(fn):
    fn_src = inspect.getsource(fn)
    func_tree = ast.parse(textwrap.dedent(fn_src))
    fn_body = func_tree.body[0]
    func_args = fn_body.args.args
    return extract_arg_name_order_from_ast(func_args)


def extract_sensitivity_from_dec(deco_list, fn_name):
    if len(deco_list) == 0:
        return CodeBlockType.Combinational, []
    else:
        assert len(deco_list) == 1, "{0} is not called with multiple decorators blocks".format(fn_name)
        call_obj = deco_list[0]
        if isinstance(call_obj, ast.Call):
            call_name = call_obj.func.id
        else:
            assert isinstance(call_obj, ast.Name), "Unrecognized  function decorator {0}".format(call_obj)
            call_name = call_obj.id
        if call_name == "always_comb":
            return CodeBlockType.Combinational, []
        elif call_name == "initial":
            return CodeBlockType.Initial, []
        elif call_name == "always_latch":
            return CodeBlockType.Latch, []
        else:
            assert call_name == "always_ff", "Unrecognized function decorator {0}".format(call_name)
        blk_type = CodeBlockType.Sequential
        raw_sensitivity = call_obj.args
        result = []
        # TODO: fix me. the frame num calculation is a hack
        local = get_frame_local(4)
        for entry in raw_sensitivity:
            assert len(entry.elts) == 2
            edge_node, signal_name_node = entry.elts
            if isinstance(edge_node, ast.Name):
                edge_type = edge_node.id
            else:
                edge_type = edge_node.attr
            edge_type = edge_type.capitalize()
            if isinstance(signal_name_node, ast.Name):
                name = signal_name_node.id
                assert name in local, "{0} not found".format(name)
                n = eval(name, local)
                assert isinstance(n, _kratos.Var), "{0} is not a variable".format(name)
                signal_name = n
            elif isinstance(signal_name_node, ast.Attribute):
                # need to eval the actual name
                n = eval(astor.to_source(signal_name_node), local)
                assert isinstance(n, _kratos.Var), "{0} is not a variable".format(signal_name_node)
                signal_name = n
            else:
                signal_name = signal_name_node.s
            result.append((edge_type, signal_name))
        return blk_type, result


def get_ln(fn):
    info = inspect.getsourcelines(fn)
    return info[1]


def get_fn(fn):
    return os.path.abspath(inspect.getsourcefile(fn))



from kratos import Generator, PortDirection, PortType, always_ff, \
    verilog, is_valid_verilog, VarException, StmtException, \
    PackedStruct, Attribute, ext, posedge, PortBundle, const, comment, \
    enable_runtime_debug, always_comb
from _kratos.passes import uniquify_generators, hash_generators_parallel
import os
import tempfile
import shutil
import pytest
import platform
import kratos

no_verilator = shutil.which("verilator") is None
is_windows = platform.system() == "Windows"


class PassThroughMod(Generator):
    def __init__(self, is_clone: bool = False):
        super().__init__("mod1", True, is_clone)
        self.in_ = self.input("in", 1)
        self.out_ = self.output("out", 1)
        self.wire(self.out_, self.in_)


class PassThroughTop(Generator):
    def __init__(self):
        super().__init__("top", True)

        self.input("in", 1)
        self.output("out", 1)

        pass_through = PassThroughMod()
        self.add_child("pass", pass_through)
        self.wire(self["pass"].ports["in"], self.ports["in"], )

        self.wire(self.ports.out, self["pass"].ports.out)


def test_generator():
    mods = []
    for i in range(10):
        mod = Generator("mod1")
        mod.instance_name = f"a{i}"
        mods.append(mod)

    for idx, mod in enumerate(mods):
        assert mod.name == "mod1"
        mod.name = f"new_mod{idx}"
        assert mod.name == f"new_mod{idx}"


class AsyncReg(Generator):
    def __init__(self, width, debug=False):
        super().__init__("register", debug)

        # define inputs and outputs
        self._in = self.input("in", width)
        self._out = self.output("out", width)
        self._clk = self.clock("clk")
        self._rst = self.reset("rst")
        self._val = self.var("val", width)

        # add combination and sequential blocks
        self.add_always(self.seq_code_block)

        self.add_always(self.comb_code_block)

    @always_ff((posedge, "clk"), (posedge, "rst"))
    def seq_code_block(self):
        if self._rst:
            self._val = 0
        else:
            self._val = self._in

    @always_comb
    def comb_code_block(self):
        self._out = self._val


def test_async_reg(check_gold):
    reg_width = 16
    reg = AsyncReg(reg_width)
    check_gold(reg, "test_async_reg")
    assert "in" in reg.ports
    assert "val" in reg.vars


def test_module_unique():
    reg1 = AsyncReg(16)
    reg2 = AsyncReg(1)
    reg2.instance_name = "test"
    parent = Generator("top")
    parent.add_child("reg1", reg1)
    parent.add_child("reg2", reg2)

    hash_generators_parallel(parent.internal_generator)
    c = Generator.get_context()
    reg1_hash = c.get_hash(reg1.internal_generator)
    reg2_hash = c.get_hash(reg2.internal_generator)
    assert reg1_hash != reg2_hash

    uniquify_generators(parent.internal_generator)
    assert reg1.internal_generator.name != reg2.internal_generator.name
    assert reg1.name != reg2.name


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_wire_const():
    mod = Generator("test")
    out_ = mod.output("out", 1)
    mod.wire(out_, 1)

    mod_src = verilog(mod)
    assert is_valid_verilog(mod_src)


def test_else_if(check_gold):
    class ElseIf(Generator):
        def __init__(self):
            super().__init__("elseif")
            self._in0 = self.port("in0", 1, PortDirection.In)
            self._in1 = self.port("in1", 1, PortDirection.In)
            self._out = self.port("out", 1, PortDirection.Out)

            self.add_always(self.else_if)

        @always_comb
        def else_if(self):
            if self._in0 == const(1, 1):
                self._out = 1
            elif self._in1 == const(1, 1):
                self._out = 0
            else:
                self._out = 1

    mod = ElseIf()
    check_gold(mod, "test_else_if")


def test_mod_instantiation(check_gold):
    mod = PassThroughTop()
    # turn off pass through module optimization since it will remove
    # mod2 completely
    check_gold(mod, "test_mod_instantiation", optimize_passthrough=False)


def test_external_module():
    src_file = os.path.join(os.path.dirname(__file__), "vectors", "module1.sv")
    mod = Generator.from_verilog("module1", src_file, [], {})
    assert mod.internal_generator.external_filename() == src_file
    assert mod.name == "module1"
    assert mod.internal_generator.external()
    hash_generators_parallel(mod.internal_generator)
    c = Generator.get_context()
    assert c.has_hash(mod.internal_generator)
    assert c.get_hash(mod.internal_generator) != 0


class ForLoopModule(Generator):
    def __init__(self, num_var: int, debug: bool, loop_input: bool):
        super().__init__("mod", debug)
        self.num_var = num_var
        self.loop_input = loop_input
        if self.loop_input:
            self.inputs = []
            for i in range(num_var):
                self.inputs.append(self.port(f"in{i}", 1, PortDirection.In))
            self.out = self.output("out", num_var)
        else:
            self.out = self.output("out", num_var, size=num_var)

        self.add_always(self.code_block)

    @always_comb
    def code_block(self):
        for i in range(self.num_var):
            if self.loop_input:
                self.out[i] = self.inputs[i]
            else:
                self.out[i] = i


def test_for_loop_debug(check_gold):
    mod = ForLoopModule(4, True, True)
    check_gold(mod, "test_for_loop")


def test_for_loop_no_debug_unroll(check_gold):
    # because self.inputs is not a kratos variable, it should unroll
    mod = ForLoopModule(4, False, True)
    check_gold(mod, "test_for_loop")


def test_for_loop_no_unroll(check_gold):
    mod = ForLoopModule(4, False, False)
    check_gold(mod, "test_for_loop_no_unroll")


def test_for_bit_index():
    class Mod(Generator):
        def __init__(self, num_loop):
            super().__init__("mod")
            self.out_ = self.output("out", num_loop)
            self.num_loop = num_loop

            self.add_always(self.code)

        @always_comb
        def code(self):
            for i in range(self.num_loop):
                self.out_[i] = 1

    mod = Mod(4)
    src = verilog(mod)["mod"]
    assert "out[2'(i)] = 1'h1;" in src


def test_switch(check_gold):
    class Switch(Generator):
        def __init__(self):
            super().__init__("switch_test")

            self._in = self.port("in", 3, PortDirection.In)
            self._out = self.port("out", 3, PortDirection.Out)

            self.add_always(self.logic)

        @always_comb
        def logic(self):
            if self._in == const(0, 3):
                self._out = 0
            elif self._in == const(1, 3):
                self._out = 1
            else:
                self._out = 2

    mod = Switch()
    check_gold(mod, "test_switch")


def test_pass_through(check_gold):
    mod = PassThroughTop()
    # turn off pass through module optimization since it will remove
    # mod2 completely
    check_gold(mod, "test_pass_through", optimize_passthrough=True)


def test_nested_if(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod1")
            self.in_ = self.port("in", 2, PortDirection.In)
            self.out_ = self.port("out", 2, PortDirection.Out)

            self.add_always(self.nested_if)

        @always_comb
        def nested_if(self):
            if self.in_ < const(1, 2):
                if self.in_ < const(2, 2):
                    self.out_ = 1
                else:
                    self.out_ = 3
            else:
                self.out_ = 2

    mod = Mod()
    check_gold(mod, "test_nested_if")


def test_fanout_mod_inst(check_gold):
    class Mod2(Generator):
        def __init__(self):
            super().__init__("mod2")
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 1, PortDirection.Out)

            self.mod_1 = PassThroughMod()
            self.mod_2 = PassThroughMod()

            self.add_child("mod1", self.mod_1)
            self.add_child("mod3", self.mod_2)

            self.wire(self.in_, self.mod_1.in_)
            self.wire(self.in_, self.mod_2.in_)

            self.add_always(self.code)

        @always_comb
        def code(self):
            self.out_ = self.mod_1.out_ + self.mod_2.out_

    mod = Mod2()
    check_gold(mod, "test_fanout_mod_inst", optimize_passthrough=False)
    Generator.clear_context()
    mod2 = Mod2()
    check_gold(mod2, "test_fanout_mod_inst_passthrough")


def test_debug():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_1 = self.port("out1", 1, PortDirection.Out)
            self.out_2 = self.port("out2", 1, PortDirection.Out)

            self.wire(self.out_1, self.in_)

            self.add_always(self.code)

        @always_comb
        def code(self):
            self.out_2 = self.in_

    mod = Mod()
    mod_src, mod_debug = verilog(mod, debug_fn_ln=True)
    src_mapping = mod_debug["mod1"]
    assert len(src_mapping) == 7
    verilog_lines = mod_src["mod1"].split("\n")
    verilog_ln = 0
    for ln, line in enumerate(verilog_lines):
        if "assign out1 = in;" in line:
            verilog_ln = ln + 1
            break
    fn, ln = src_mapping[verilog_ln][0]
    with open(fn) as f:
        python_lns = f.readlines()
    assert "self.wire(self.out_1, self.in_)" in python_lns[ln - 1]


def test_illegal_assignment_width():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 4, PortDirection.Out)

            self.add_always(self.code)

        @always_comb
        def code(self):
            if self.in_ == const(1, 1):
                self.out_ = const(1, 4)
            else:
                self.out_ = const(100, 10)

    try:
        Mod()
        assert False
    except VarException as ex:
        print(ex)
        assert True


def test_illegal_assignment_blocking():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 1, PortDirection.Out)
            self.clk_ = self.port("clk", 1, PortDirection.In, PortType.Clock)

            self.wire(self.out_, self.in_)

            self.add_always(self.code)

        @always_ff((posedge, "clk"))
        def code(self):
            self.out_ = 1

    try:
        mod = Mod()
        verilog(mod)
        assert False
    except StmtException as ex:
        print(ex)
        assert True


def test_data_if(check_gold):
    class Mod(Generator):
        def __init__(self, bool_flag):
            super().__init__("mod1")
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 1, PortDirection.Out)
            self.bool_flag = bool_flag

            self.add_always(self.code)

        @always_comb
        def code(self):
            if self.bool_flag:
                if self.in_ == const(1, 1):
                    self.out_ = 1
                else:
                    self.out_ = 0
            else:
                if self.in_ == const(0, 1):
                    self.out_ = 0
                else:
                    self.out_ = 1

    mod = Mod(True)
    check_gold(mod, "test_data_if_true")
    # need to clear the context, otherwise it will be a different module name
    Generator.clear_context()
    mod = Mod(False)
    check_gold(mod, "test_data_if_false")


def test_static_eval_for_loop(check_gold):
    class Mod(Generator):
        def __init__(self, num_loop):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", num_loop, PortDirection.Out)
            self.num_loop = num_loop

            self.add_always(self.code)

        @always_comb
        def code(self):
            if self.in_ == const(1, 1):
                for i in range(self.num_loop):
                    self.out_[i] = 1
            else:
                for i in range(self.num_loop):
                    self.out_[i] = 0

    loop = 4
    mod = Mod(loop)
    mod_src, mod_debug = verilog(mod, debug_fn_ln=True)
    mod_mapping = mod_debug["mod1"]
    lines = list(mod_mapping.keys())
    lines.sort()
    for ii in range(len(mod_mapping) - loop, len(mod_mapping) - 1):
        assert mod_mapping[lines[-1]][-1] == mod_mapping[lines[ii]][-1]
    Generator.clear_context()
    check_gold(mod, "test_static_eval_for_loop")


def test_const_port(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")

            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 2, PortDirection.Out)

            self.child = PassThroughMod()
            self.add_child("child", self.child)
            self.wire(self.child.in_, const(0, 1))
            self.wire(self.out_[0], self.child.out_)
            self.wire(self.out_[1], self.in_)

    mod = Mod()
    check_gold(mod, "test_const_port", optimize_passthrough=False)


def test_create(check_gold):
    class Mod(Generator):
        def __init__(self, width, is_clone=False):
            super().__init__(f"mod_{width}", is_clone=is_clone)

            self.in_ = self.port("in", width, PortDirection.In)
            self.out_ = self.port("out", width, PortDirection.Out)
            self.wire(self.out_, self.in_)

    mod1 = Mod.create(width=1)
    mod2 = Mod.create(width=2)
    mod3 = Mod.create(width=1)

    assert not mod1.is_cloned
    assert not mod2.is_cloned
    assert mod3.is_cloned
    assert mod3.def_instance == mod1

    # modify mod 3
    mod3.initialize_clone()
    mod3.in_.width = 3
    mod3.out_.width = 3
    assert not mod3.is_cloned
    check_gold(mod3, "test_create")


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_clone():
    class Mod2(Generator):
        def __init__(self):
            super().__init__("mod2")
            self.in_ = self.input("in", 2)
            self.out_ = self.output("out", 2)

            self.child1 = PassThroughMod.clone()
            self.child2 = PassThroughMod.clone()
            self.add_child("child1", self.child1)
            self.add_child("child2", self.child2)

            self.add_always(self.code)

        @always_comb
        def code(self):
            self.child1.ports["in"] = self.in_[0]
            self.child2.ports["in"] = self.in_[1]

            self.out_[0] = self.child1.ports.out
            self.out_[1] = self.child2.ports.out

    mod = Mod2()
    assert not mod.child1.is_cloned
    assert mod.child2.is_cloned
    mod_src = verilog(mod, optimize_fanout=False, optimize_passthrough=False)
    assert is_valid_verilog(mod_src)


def test_packed_struct(check_gold, check_file):
    struct = PackedStruct("config_data", [("read", 16),
                                          ("data", 16)])

    class Mod(Generator):
        def __init__(self, debug=False):
            super().__init__("mod", debug=debug)
            self.input("in", struct)
            self.output("out", struct)
            v = self.var_packed("v", struct)
            self.wire(self.ports["out"], self.ports["in"])
            self.wire(v, self.ports["in"])

            # tests
            # assert v.width == 16 + 16
            assert "read" in v.member_names()
            assert "data" in v.member_names()
            assert v["read"].width == 16

    mod = Mod()
    check_gold(mod, "test_packed_struct", optimize_passthrough=False)
    Generator.clear_context()
    mod = Mod(True)
    mod.name = "packed_struct"
    with tempfile.TemporaryDirectory() as temp:
        verilog(mod, output_dir=temp, debug_fn_ln=True)
        mod_file = os.path.join(temp, "packed_struct.sv")
        def_file = os.path.join(temp, "packed_struct_pkg.svh")
        import json
        # json is correct
        if is_windows:
            pytest.skip("Windows new line problem")
        with open(os.path.join(temp, "packed_struct.sv.debug")) as f:
            json.load(f)
        check_file(mod_file, "packed_struct_pkg.sv")
        check_file(def_file, "packed_struct_pkg.svh")


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_more_debug1():
    mod = PassThroughTop()
    mod_src, debug_info = verilog(mod, debug_fn_ln=True)
    debug = debug_info["top"]
    assert is_valid_verilog(mod_src)
    assert len(debug) > 3


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_more_debug2():
    class Top(Generator):
        def __init__(self):
            super().__init__("top", True)

            self.port("in", 1, PortDirection.In)
            self.port("out", 1, PortDirection.Out)

            pass_through = PassThroughMod()
            self.add_child("pass", pass_through)
            self.wire(self["pass"].ports["in"], self.ports["in"], )

            self.add_always(self.code_block)

        @always_comb
        def code_block(self):
            self.ports.out = self["pass"].ports.out

    mod = Top()
    mod_src, debug_info = verilog(mod, debug_fn_ln=True)
    debug = debug_info["top"]
    assert is_valid_verilog(mod_src)
    assert len(debug) > 3


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_verilog_file():
    mod = PassThroughTop()
    with tempfile.TemporaryDirectory() as tempdir:
        filename = os.path.join(tempdir, "mod.sv")
        verilog(mod, filename=filename, debug_fn_ln=True, optimize_passthrough=False)
        with open(filename) as f:
            src = f.read()
            assert is_valid_verilog(src)


def test_wire_merge(check_gold):
    class TestModule(Generator):
        def __init__(self, width):
            super().__init__("Test")
            self.port("in", width, PortDirection.In)
            self.port("out", width, PortDirection.Out)

            for i in range(width):
                self.wire(self.ports["out"][i], self.ports["in"][i])

    mod = TestModule(4)
    check_gold(mod, "test_wire_merge")


def test_remove_child():
    top = PassThroughTop()
    child = top["pass"]
    assert child in top
    top.remove_child_generator(child)
    assert child not in top
    # top should be empty now
    assert top.stmts_count == 0


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_syntax_sugar():
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 1)
    in_ = mod.input("in", 1)
    comb = mod.combinational()
    comb.if_(in_ == 1).then_(out_(0)).else_(out_(1))

    mod_src = verilog(mod)
    is_valid_verilog(mod_src)

    Generator.clear_context()
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 1)
    in_ = mod.input("in", 1)
    comb = mod.combinational()
    comb.switch_(in_).case_(1, out_(1)).case_(0, out_(0))

    mod_src = verilog(mod)
    is_valid_verilog(mod_src)


@pytest.mark.skipif(no_verilator, reason="verilator not available")
def test_zero_ext():
    mod = Generator("mod", debug=True)
    out_ = mod.port("out", 2, PortDirection.Out)
    in_ = mod.port("in", 1, PortDirection.In)
    mod.wire(out_, ext(in_, 2))

    mod_src = verilog(mod)
    is_valid_verilog(mod_src)


def test_port_array(check_gold):
    mod = Generator("mod", True)
    in_ = mod.port("in", 2, PortDirection.In, size=2)
    out1 = mod.port("out1", 2, PortDirection.Out, size=2)
    out2 = mod.port("out2", 2, PortDirection.Out, size=2)
    mod.wire(out1, in_)
    mod.wire(out2[0][0], in_[0][1])
    mod.wire(out2[0][1], in_[0][0])
    mod.wire(out2[1], in_[1])

    check_gold(mod, "test_port_array")


def test_simple_pipeline(check_gold):
    mod = PassThroughMod()
    # add a clock
    mod.clock("clk")
    attr = Attribute()
    attr.type_str = "pipeline"
    attr.value_str = "2"
    mod.add_attribute(attr)

    check_gold(mod, "test_simple_pipeline", insert_pipeline_stages=True)


def test_replace(check_gold):
    mod = PassThroughTop()

    class Mod(Generator):
        def __init__(self):
            super().__init__("test", True)
            self.in_ = self.input("in", 1)
            self.out_ = self.output("out", 1)
            self.wire(self.out_, self.in_)

    child = Mod()
    mod.replace("pass", child)
    check_gold(mod, "test_replace", optimize_passthrough=False)


def test_local_function(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("test")
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)
            clk = self.clock("clk")

            @always_ff((posedge, "clk"))
            def code_block():
                if clk:
                    out_ = in_

            self.add_always(code_block)

    mod = Mod()
    check_gold(mod, "test_local_function")


def test_reg_next(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("test")
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)
            try:
                a = self.reg_next("var1", in_)
                assert False
            except AssertionError:
                assert True
            # add a clock
            self.clock("clk")
            a = self.reg_next("a", in_)
            b = self.reg_next("b", a)
            self.wire(out_, b)

    mod = Mod()
    check_gold(mod, "test_reg_next")


def test_reg_init(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("test")
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)
            # add a clock
            self.clock("clk")
            # add a reset
            self.reset("rst")
            a = self.reg_init("a", in_)
            b = self.reg_init("b", a)
            self.wire(out_, b)

    mod = Mod()
    check_gold(mod, "test_reg_init")


def test_reg_enable(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("test")
            in_ = self.input("in", 1)
            en_ = self.input("en", 1)
            out_ = self.output("out", 1)
            # add a clock
            self.clock("clk")
            a = self.reg_enable("a", in_, en_)
            b = self.reg_enable("b", a, en_)
            self.wire(out_, b)

    mod = Mod()
    check_gold(mod, "test_reg_enable")


def test_ternary(check_gold):
    from kratos import mux

    class Mod(Generator):
        def __init__(self):
            super().__init__("test")
            in1 = self.input("in1", 1)
            in2 = self.input("in2", 1)
            in3 = self.input("in3", 1)
            out = self.output("out", 1)

            self.wire(out, mux(in1, in2, in3))

    mod = Mod()
    check_gold(mod, "test_ternary")


def test_bundle(check_gold):
    class Test(PortBundle):
        def __init__(self):
            super().__init__(True)
            self.input("a", 1)
            self.output("b", 1)

    class Mod(Generator):
        def __init__(self):
            super().__init__("test_bundle", True)

            p = self.port_bundle("in_port", Test())
            self.port_bundle("out_port", Test().flip())

            self.wire(p.a, self.ports.out_port.a)
            self.wire(self.ports.in_port.b, self.ports.out_port.b)

    mod = Mod()
    check_gold(mod, "test_bundle")


def test_bundle_pack():
    class Test(PortBundle):
        def __init__(self):
            super().__init__(True)
            self.input("a", 1)
            self.input("b", 1)

    class Mod1(Generator):
        def __init__(self):
            super().__init__("mod1")

            p = self.port_bundle("p", Test())
            self.p = p
            self.a = self.output("a", 1)
            self.b = self.output("b", 1)
            self.wire(self.a, p["a"])
            self.wire(self.b, p["b"])

    class Mod2(Generator):
        def __init__(self):
            super().__init__("mod2")

            p = self.port_bundle("p", Test().flip())
            self.p = p
            self.a = self.input("a", 1)
            self.b = self.input("b", 1)
            self.wire(p["a"], self.a)
            self.wire(p["b"], self.b)

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            in_ = self.input("a", 1)
            out_ = self.output("b", 1)
            mod1 = Mod1()
            mod2 = Mod2()

            self.add_child("m1", mod1)
            self.add_child("m2", mod2)

            self.wire(mod1.p.a, in_)
            self.wire(mod1.p.b, in_)
            self.wire(mod2.a, in_)
            self.wire(mod2.b, in_)
            self.add_stmt(out_.assign(mod1.a + mod2.p.a))

    mod = Mod()
    mod_src = verilog(mod, optimize_passthrough=False)
    # assert is_valid_verilog(mod_src)


def test_named_block(check_gold):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 1)
    in_ = mod.port("in", 1, PortDirection.In)
    comb = mod.combinational()
    if_ = comb.if_(in_ == 1)
    if_.then_(out_(0)).else_(out_(1))
    mod.mark_stmt("TEST", if_.then_body())

    @always_comb
    def code():
        out_ = 1

    mod.add_always(code, label="TEST2")

    check_gold(mod, "test_named_block", check_multiple_driver=False)


def test_enum(check_gold):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 1)
    in_ = mod.input("in", 1)
    mod.enum("color", {"red": 1, "green": 2}, 2)
    mod.wire(out_, in_)

    check_gold(mod, "test_enum")


def setup_fsm(fsm, out_, in_):
    # add outputs
    fsm.output(out_)
    # add states
    red = fsm.add_state("Red")
    blue = fsm.add_state("Blue")
    # set the state transition
    red.next(red, in_ == 0)
    red.next(blue, in_ == 1)
    blue.next(red, in_ == 1)
    # fill in outputs
    red.output(out_, 2)
    blue.output(out_, 1)
    # set the start case
    fsm.set_start_state("Red")


def test_fsm(check_gold, check_file):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    # setup FSM
    setup_fsm(fsm, out_, in_)

    check_gold(mod, "test_fsm", optimize_if=False)
    # output fsm graph
    dot = fsm.dot_graph()
    check_file(dot, "test_fsm.dot")
    csv = fsm.output_table()
    check_file(csv, "test_fsm.csv")


def test_fsm_mealy(check_gold):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    # setup FSM
    setup_fsm(fsm, out_, in_)
    # use mealy
    fsm.is_moore = False
    check_gold(mod, "test_fsm_mealy", optimize_if=False)


def test_function(check_gold):
    from kratos.func import function

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 2)
            self._out2 = self.output("out2", 2)

            self.add_always(self.code)

        @function
        def update_out(self, value, predicate):
            self._out2 = value
            if predicate:
                return value + self._in
            else:
                return value

        @always_comb
        def code(self):
            # because the types are inferred
            # implicit const conversion doesn't work here
            self._out = self.update_out(self._in, const(1, 1))

    mod = Mod()
    check_gold(mod, "test_function")


def test_function_missing_return():
    from kratos.func import function

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 2)

            self.add_always(self.code)

        @function
        def update_out(self, value, predicate):
            self._out = value
            if predicate:
                return value + self._in

        @always_comb
        def code(self):
            # because the types are inferred
            # implicit const conversion doesn't work here
            self._out = self.update_out(self._in, const(1, 1))

    mod = Mod()
    try:
        verilog(mod)
        assert False
    except StmtException:
        assert True


def test_reg_file(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            self._in = self.input("in", 4)
            self._out = self.output("out", 4)
            self.waddr = self.input("warr", 2)
            self.raddr = self.input("raddr", 2)
            self.reg = self.var("reg_file", 4, size=4)

            self.add_always(self.code)

        @always_comb
        def code(self):
            self.reg[self.waddr] = self._in
            self._out = self.reg[self.raddr]

    mod = Mod()
    check_gold(mod, "test_reg_file")


def test_comment(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            self._in = self.input("in", 1)
            self._out = self.output("out", 1)
            self._out2 = self.output("out2", 1)
            self._out3 = self.output("out3", 1)
            self.var = self.var("v", 1)

            child = PassThroughMod()
            self.add_child("child", child, "This is a comment")
            self.wire(self._in, child.in_)
            self.wire(self._out, child.out_)
            self.wire(self._out2, self.var)

            self._in.comment = "Input port"
            self.var.comment = "variable comment"
            self.add_always(self.code)

        @always_comb
        def code(self):
            comment("Another comment")
            self._out3 = self._in

    mod = Mod()
    check_gold(mod, "test_comment", optimize_passthrough=False)


def test_packed_array(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            self._in = self.input("in", 4, size=6, packed=True)
            self._out = self.output("out", 4, size=6, packed=True)
            self.wire(self._in, self._out)

    mod = Mod()
    check_gold(mod, "test_packed_array")


def test_rename():
    mod = PassThroughTop()
    child = mod["pass"]
    child.instance_name = "test"
    assert child == mod["test"]
    assert mod.internal_generator.has_child_generator("test")
    child.name = "test2"
    assert mod["test"].internal_generator.name == "test2"


def test_c_dpi_function(check_gold):
    from kratos.func import dpi_function

    @dpi_function(8)
    def add(arg0, arg1):
        pass

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 8)

            self.add_always(self.code)

        @always_comb
        def code(self):
            # because the types are inferred
            # implicit const conversion doesn't work here
            self._out = add(self._in, const(1, 2))

    mod = Mod()
    # once it's turned off, user has to handle sv logic themselves in the
    # c interface
    check_gold(mod, "test_dpi", int_dpi_interface=False)


def test_nested_fsm(check_gold, check_file):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    setup_fsm(fsm, out_, in_)
    second_fsm = mod.add_fsm("HSV")
    fsm.add_child_fsm(second_fsm)
    idle = second_fsm.add_state("idle")
    idle.next(fsm["Red"], in_ == 0)
    fsm["Red"].next(idle, in_ == 2)
    second_fsm.output(out_)
    idle.output(out_, 2)

    dot = fsm.dot_graph()
    check_file(dot, "test_nested_fsm.dot")
    check_gold(mod, "test_nested_fsm", optimize_if=False)


def test_symbol_table():
    from kratos.debug import extract_symbol_table
    mod = AsyncReg(16, True)
    verilog(mod)
    table, _ = extract_symbol_table(mod)
    assert len(table) == 1
    assert len(table[mod]) == 5


def test_cast():
    from kratos.util import clock

    class Mod1(Generator):
        def __init__(self):
            super().__init__("mod")
            self.v = self.var("v", 1)
            self.out = self.output("out", 1)
            self.add_always(self.code)

        @always_ff((posedge, "v"))
        def code(self):
            self.out = 1

    mod = Mod1()
    try:
        verilog(mod)
        assert False
    except StmtException:
        assert True

    class Mod2(Generator):
        def __init__(self):
            super().__init__("mod")
            self.v = self.var("v", 1)
            clk = clock(self.v)
            # only procedural allowed
            seq = self.sequential((posedge, clock(self.v)))
            seq.add_stmt(self.output("out", 1).assign(const(1, 1)))

            @always_ff((posedge, clk))
            def fn():
                self.v = 1

            self.add_always(fn)

    mod = Mod2()
    verilog(mod)


def test_async_no_latch():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            clk = self.clock("clk")
            rst = self.reset("rst")
            cen = self.input("cen", 1, PortType.ClockEnable)
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)

            @always_ff((posedge, "clk"), (posedge, "rst"))
            def code():
                if rst:
                    out_ = 0
                elif cen:
                    out_ = in_

            self.add_always(code)

    mod = Mod()
    verilog(mod)


def test_async_latch():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            clk = self.clock("clk")
            rst = self.reset("rst")
            cen = self.input("cen", 1, PortType.ClockEnable)
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)

            @always_ff((posedge, "clk"), (posedge, "rst"))
            def code():
                if rst:
                    out_ = 0

            self.add_always(code)

    mod = Mod()
    try:
        verilog(mod)
        assert False
    except StmtException:
        assert True


def test_param(check_gold):
    mod = Generator("mod", True)
    param = mod.parameter("P", 4, 4)
    param2 = mod.parameter("P2", 4, 4)
    param3 = mod.parameter("P3", 4)  # P3 doesn't have init or value
    in_ = mod.input("in", param)
    out = mod.output("out", param2)
    var = mod.var("v", param)
    mod.wire(var, in_)
    mod.wire(out, var * 2)

    check_gold(mod, "test_param")
    param.value = 2
    try:
        verilog(mod)
        assert False
    except VarException:
        assert True
    param.value = 4
    verilog(mod)


def test_nested_param(check_gold):
    parent = Generator("parent")
    param1 = parent.parameter("P", 4, 4)
    # use normal variable to see how decoupling works
    in1 = parent.var("in", param1)
    out1 = parent.var("out", param1)

    child = Generator("child")
    param2 = child.parameter("P2", 4, 4)
    in2 = child.input("in", param2)
    out2 = child.output("out", param2)
    child.wire(out2, in2)

    parent.add_child("mod", child)
    parent.wire(in1, in2)
    parent.wire(out1, out2)

    # change the value
    param2.value = param1
    # test the value propagation
    param1.value = 2
    check_gold(parent, "test_nested_param", optimize_passthrough=False)


def hash_param_width():
    class Mod(Generator):
        def __init__(self, value):
            super().__init__("mod")
            p = self.parameter("P", 2, 2)
            out = self.output("out", p)
            self.wire(out, p)
            p.value = value

    mod1 = Mod(1)
    mod2 = Mod(2)
    parent = Generator("parent")
    parent.add_child("mod1", mod1)
    parent.add_child("mod2", mod2)

    hash_generators_parallel(parent.internal_generator)
    c = Generator.get_context()
    hash1 = c.get_hash(mod1.internal_generator)
    hash2 = c.get_hash(mod2.internal_generator)
    assert hash1 == hash2


def test_long_statement(check_gold):
    from kratos import concat
    mod = Generator("mod")
    a = mod.input("this_is_a_long_name", 1)
    num_concat = 6
    b = mod.output("out", num_concat)
    mod.wire(b, concat(*[a for _ in range(num_concat)]))
    check_gold(mod, "test_long_statement")


def test_create_stub(check_file):
    from kratos import create_stub
    mod = Generator("mod")
    mod.input("a", 16, size=16)
    mod.output("b", 1)
    mod.input("c", 16)
    mod.var("d", 1)  # this should not be generated
    check_file(create_stub(mod), "test_create_stub.sv")


def test_fsm_state(check_gold):
    from kratos import negedge
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    clk = mod.clock("clk")
    rst = mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color", reset_high=False)
    # setup FSM
    setup_fsm(fsm, out_, in_)
    # realize fsm now
    fsm.realize()
    current_state = fsm.current_state
    # create an enum var based on current_state
    state_enum = current_state.enum_type()
    # can create a variable from the enum definition
    # notice that variable s will be optimized away if not used
    s = mod.enum_var("s", state_enum)
    c = mod.var("counter", 1)

    @always_ff((posedge, clk), (negedge, rst))
    def counter():
        if rst.r_not():
            c = 0
        elif current_state == state_enum.Red:
            c = c + 1

    mod.add_always(counter)
    check_gold(mod, "test_fsm_state", optimize_if=False)


def test_not_if(check_gold):
    mod = Generator("mod")
    a = mod.output("a", 1)
    clk = mod.clock("clk")
    rst = mod.reset("rst", is_async=False)  # synchronous reset

    @always_ff((posedge, clk))
    def code():
        if not rst:
            a = 1
        else:
            a = 0

    mod.add_always(code)
    check_gold(mod, "test_not_if")


def test_exception(check_gold):
    mod = Generator("mod", True)
    in_ = mod.input("in", 1)

    @always_comb
    def code():
        if in_ == 1:
            raise Exception()

    mod.add_always(code)
    check_gold(mod, "test_exception")


def test_enum_port(check_gold):
    from kratos import enum, has_enum
    mod = Generator("mod")
    enum_ = enum("State", ["IDLE", "WAIT", "WORK"])
    assert has_enum("State")
    in_ = mod.input("in", enum_)
    out = mod.output("out", enum_)
    mod.wire(out, in_)
    check_gold(mod, "test_enum_port")
    Generator.clear_context()
    mod = Generator("mod")
    enum_ = enum("State", [("IDLE", 0), ("WAIT", 1), ("WORK", 2)])
    in_ = mod.input("in", enum_)
    out = mod.output("out", enum_)
    mod.wire(out, in_)
    check_gold(mod, "test_enum_port")


def test_interface_modport_local(check_gold):
    from kratos import Interface
    mod = Generator("mod")

    class ConfigInterface(Interface):
        def __init__(self):
            Interface.__init__(self, "Config")
            self.var("read", 1, 1)
            self.var("write", 1, 1)
            # create modport
            self.slave = self.modport("Slave")
            self.slave.set_input("read")
            self.slave.set_output("write")

    interface = ConfigInterface()
    i1 = mod.interface(interface, "bus")
    child = Generator("child")
    mod.add_child("child", child)
    # set it to True for port interface
    child_bus = child.interface(interface.slave, "bus_s", True)
    # child logic with the interface ports
    child.wire(child_bus.write, child_bus.read)
    # wire the parent to child
    mod.wire(i1.Slave, child_bus)

    check_gold(mod, "test_interface_modport_local", optimize_passthrough=False)


def test_interface_port_wiring(check_gold):
    from kratos import Interface
    mod = Generator("mod")

    class ConfigInterface(Interface):
        def __init__(self):
            Interface.__init__(self, "Config")
            self.input("read", 1, 1)
            self.output("write", 1, 1)

    interface = ConfigInterface()
    i1 = mod.interface(interface, "bus")
    # wire local variables to the interface
    read = mod.var("read", 1)
    write = mod.var("write", 1)
    mod.wire(i1.read, read)
    mod.wire(write, i1.write)

    # create child module and use the interface as port
    child = Generator("child")
    mod.add_child("child", child)
    i2 = child.interface(interface, "bus2", True)
    child.wire(i2.write, i2.read)

    # wire the interface
    mod.wire(i1, i2)
    check_gold(mod, "test_interface_port_wiring", optimize_passthrough=False)


def test_track_generated_definition():
    mod = PassThroughMod()
    with tempfile.TemporaryDirectory() as tempdir:
        filename = os.path.join(tempdir, "test.sv")
        verilog(mod, filename=filename, track_generated_definition=True)
        t1 = os.path.getmtime(filename)
        verilog(mod, filename=filename, track_generated_definition=True)
        with open(filename) as f:
            content = f.read()
        assert content == ""  # empty since no content is generated


def test_call_always():
    @always_comb
    def code():
        pass

    try:
        code()
        assert False
    except SyntaxError:
        assert True


def test_wrapper_flatten_generator(check_gold):
    mod = Generator("mod")
    a = mod.input("a", 4, size=[3, 2])
    b = mod.output("b", 4, size=[3, 2])
    mod.wire(a, b)
    from _kratos import create_wrapper_flatten
    wrapper = create_wrapper_flatten(mod.internal_generator, "wrapper")
    wrapper = Generator(wrapper.name, internal_generator=wrapper)
    check_gold(wrapper, gold_name="test_wrapper_flatten_generator",
               optimize_passthrough=False)


def test_register_extraction():
    mod = Generator("mod")
    a = mod.input("a", 4)
    b = mod.output("b", 4)
    rst = mod.reset("rst")
    clk = mod.clock("clk")

    @always_ff((posedge, "clk"), (posedge, "rst"))
    def code():
        if rst:
            b = 0
        else:
            b = a + b

    mod.add_always(code)

    import _kratos
    regs = _kratos.passes.extract_register_names(mod.internal_generator)
    assert len(regs) == 1
    assert "mod.b" in regs


def test_iadd_transform():
    mod = Generator("mod", True)
    a = mod.var("a", 4)
    clk = mod.clock("clk")

    @always_ff((posedge, clk))
    def code():
        a += 1

    mod.add_always(code)
    src = verilog(mod)
    assert "a <= a + 4'h1;" in src["mod"]
    # make sure the line info gets passed down
    stmt = mod.get_stmt_by_index(0)[0]
    assert len(stmt.fn_name_ln) == 1


def test_if_logical_cond():
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = mod.var("b", 1)
    c = mod.var("c", 1)

    @always_comb
    def code():
        if (not a and b) or a:
            c = 1
        else:
            c = 0

    mod.add_always(code)

    @always_comb
    def illegal():
        if 1 and a:
            c = 1

    # test illegal
    try:
        mod.add_always(illegal)
        assert False
    except SyntaxError:
        assert True


def test_wiring_with_instantiation():
    mod = Generator("parent")
    child = Generator("child")
    a = mod.var("a", 1)
    mod.var("b", 1)
    mod.output("c", 1)
    child_in = child.input("in_port", 1)
    child_out1 = child.output("out_port1", 1)
    child_out2 = child.output("out_port2", 1)
    child.wire(child_out1, child_in)
    child.wire(child_out2, child_in)
    # we use the port name to instantiate the connection
    # due to Python's limitation, we can use keyword such as "in"
    # the syntax is child_port_name=parent_port
    # where parent port can be either port name or port variable
    mod.add_child("child_inst", child,
                  in_port=a, out_port1="b", out_port2="c")
    r = verilog(mod)
    assert ".out_port1(b)" in r["parent"]


def test_nested_loop():
    mod = Generator("mod")
    a = mod.var("a", 8)

    @always_comb
    def code():
        for i in range(2):
            a[i] = i
            for j in range(4):
                a[i * 4 + j] = 1

    mod.add_always(code)
    src = verilog(mod)["mod"]
    assert "a[3'(i)] = 1'(i);" in src
    assert "a[(3'(i) * 3'h4) + 3'(j)] = 1'h1;" in src
    # with debug on
    mod = Generator("mod", True)
    a = mod.var("a", 8)
    mod.add_always(code)
    src = verilog(mod)["mod"]
    assert "a[7] = 1'h1;" in src


def test_turn_off_optimization():
    class Mod(Generator):
        def __init__(self, merge):
            super().__init__("mod")
            a = self.var("a", 2)
            b = self.var("b", 2)
            c = self.var("c", 2)
            clk = self.clock("clk")

            @always_ff((posedge, clk))
            def code():
                if a == 0:
                    b = 0
                # NOTE: this is not synthesizable verilog!
                if a == 1:
                    c = 1

            self.add_always(code, merge_if_block=merge)
    src = verilog(Mod(False), check_flip_flop_always_ff=False)["mod"]
    assert "  if (a == 2'h1) begin" in src
    assert "else if" not in src
    src = verilog(Mod(True))["mod"]
    assert "else if" in src


def test_generator_property(check_gold):
    from kratos import Sequence
    mod = Generator("mod")
    a = mod.input("a", 1)
    a_tmp = mod.var("a_tmp", 1)
    b = mod.output("b", 1)
    # any sequence-based property needs a clock
    clk = mod.clock("clk")

    @always_ff((posedge, clk))
    def code():
        if (a == 1) & ~a_tmp:
            b = 0
            a_tmp = 1
        elif a_tmp:
            b = 1

    mod.add_code(code)

    seq = Sequence(a == 1)
    # if a == 1, then it imply b == 0, and the next cycle, b == 1
    seq.imply(b == 0).wait(1).imply(b == 1)
    # add this sequence to the generator
    p1 = mod.property("rule1", seq)
    p1.action = kratos.PropertyAction.Assert

    p2 = mod.property("rule2", seq)
    p2.action = kratos.PropertyAction.Assume

    p3 = mod.property("rule3", seq)
    p3.action = kratos.PropertyAction.Cover

    check_gold(mod, gold_name="test_generator_property")


def test_verilog_ln_fix():
    from kratos.func import dpi_function

    @dpi_function(8)
    def add(arg0, arg1):
        pass

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 8)
            self.add_always(self.code)

        @always_comb
        def code(self):
            self._out = add(self._in, const(1, 2))
            if self._in == 0:
                self._out = 1

    mod = Mod()
    with tempfile.TemporaryDirectory() as temp:
        filename = os.path.join(temp, "test.sv")
        src = verilog(mod, filename=filename)[0]
        content = src["mod"]

    mod_i = mod.internal_generator
    assert mod_i.verilog_ln == 2
    lines = content.split("\n")
    stmt_0 = min([i for i in range(len(lines)) if "add (" in lines[i]])
    stmt_1 = min([i for i in range(len(lines)) if "if" in lines[i]])
    # + 1 for using starting-1 line number format
    # another + 1 for DPI header offset
    assert mod.get_stmt_by_index(0)[0].verilog_ln == stmt_0 + 1 + 1
    assert mod.get_stmt_by_index(0)[1].then_body().verilog_ln == stmt_1 + 2


def transform_block_comment():
    mod = Generator("mod")

    @always_comb
    def code():
        """
        this is comment
        this is another comment
        """

    mod.add_always(code)
    src = verilog(mod)["mod"]
    assert "this is comment" in src
    assert "this is another comment" in src


def test_add_only():
    # this is for testing internal public API, advanced user only
    parent = Generator("parent")
    child = Generator("child")
    parent.add_child(child.instance_name, child, python_only=True)
    assert child not in parent


def test_always_parameter():
    mod = Generator("mod")
    a = mod.output("a", 3)

    @always_comb
    def code(i):
        a[i] = i

    @always_comb
    def code_self(self, i):
        a[i] = 0

    try:
        mod.add_always(code)
        assert False
    except AssertionError:
        pass

    try:
        mod.add_always(code, j=1)
        assert False
    except AssertionError:
        pass

    try:
        mod.add_always(code_self)
        assert False
    except AssertionError:
        pass

    mod.add_always(code, i=0)
    mod.add_always(code, i=1)
    mod.add_always(code_self, i=2)
    src = verilog(mod)["mod"]
    assert "a[0] = 1'h0;" in src
    assert "a[1] = 1'h1;" in src
    assert "a[2] = 1'h0;" in src


def test_bit_loop_slicing():
    mod = Generator("mod")
    a = mod.var("a", 1)

    @always_comb
    def code():
        for i in range(1):
            a[i] = i

    mod.add_always(code)
    src = verilog(mod)["mod"]
    assert "a = 1'h0" in src

    mod = Generator("mod")
    a = mod.var("a", 4)
    clk = mod.clock("clk")

    @always_ff((posedge, clk))
    def code():
        for i in range(3):
            if a[i]:
                a[i] = 1
    mod.add_always(code)
    src = verilog(mod)["mod"]
    assert "a[2'(i)] <= 1'h1;" in src


def test_port_cast_child():
    parent = Generator("mod")
    a = parent.var("a", 1)
    b = parent.var("b", 1)
    c = kratos.util.clock(a & b)

    child = Generator("child")
    parent.add_child("child", child)
    p = child.clock("clk")
    parent.wire(c, p)

    src = verilog(parent)["mod"]
    assert ".clk(child_clk)" in src
    assert "assign child_clk = a & b;" in src


def test_always_latch(check_gold):
    from kratos import always_latch
    mod = Generator("mod")
    a = mod.input("a", 1)
    b = mod.output("b", 1)

    @always_latch
    def code():
        if a:
            b = 1

    mod.add_always(code)
    check_gold(mod, gold_name="test_always_latch", reorder_stmts=True)


def test_scope_keyword():
    mod = Generator("mod")
    scope = mod.var("a", 1)

    @always_comb
    def code():
        pass

    try:
        mod.add_always(code)
        assert False
    except SyntaxError:
        pass


@pytest.mark.skipif(not kratos.pyast.has_format_string(), reason="format string not supported in Python")
def test_loop_format_str():
    mod = Generator("mod")
    for i in range(2):
        mod.add_child(f"a_{i}", Generator("child"))
        mod[f"a_{i}"].input("a", 1)

    @always_comb
    def code():
        for i in range(2):
            mod[f"a_{i}"].ports["a"] = 1

    mod.add_always(code)
    src = verilog(mod)["mod"]
    for j in range(2):
        assert f"child a_{j}" in src


def test_attribute_lookup():
    mod = Generator("mod")

    class NewAttr(Attribute):
        def __init__(self):
            super().__init__()
            self.value = "42"

    attr = NewAttr()
    mod.add_attribute(attr)
    result = mod.find_attribute(lambda a: isinstance(a, NewAttr))
    assert len(result) == 1
    assert result[0] == attr


def test_param_set_value_instance():
    child = Generator("child")
    parent = Generator("parent")
    child.parameter("P", 32, 0)
    parent.add_child("inst", child, P=3)
    src = verilog(parent, optimize_passthrough=False)["parent"]
    assert "child #(\n  .P(32'h3))\ninst()" in src


def test_ports_vars_iter():
    mod = Generator("mod")
    mod.input("a", 1)
    mod.var("b", 1)
    mod.output("c", 1)
    mod.parameter("P1", 1)
    mod.parameter("P2", 1)
    ports = set()
    for port_name in mod.ports:
        ports.add(port_name)
    assert len(ports) == 2
    assert "a" in ports
    assert "c" in ports
    vs = set()
    for var_name, _ in mod.vars:
        vs.add(var_name)
    assert len(vs) == 3
    params = set()
    for param_name, _ in mod.params:
        params.add(param_name)
    assert len(params) == 2


def test_param_initial_codegen():
    child1 = Generator("child1")
    child1.parameter("P", 16, initial_value=5)
    parent1 = Generator("parent1")
    parent1.add_child("inst", child1, P=2)
    src = verilog(parent1, optimize_passthrough=False)["parent1"]
    assert ".P(16'h2))\ninst();" in src

    child2 = Generator("child2")
    child2.parameter("P", 16, initial_value=5)
    parent2 = Generator("parent2")
    parent2.add_child("inst", child2, P=5)
    src = verilog(parent2, optimize_passthrough=False)["parent2"]
    assert "child2 inst()" in src


def test_port_from_def():
    mod1 = Generator("mod1")
    p1 = mod1.input("p", 4)
    mod2 = Generator("mod2")
    p2 = mod2.port_from_def(p1)
    assert str(p1) == str(p2)
    assert p1.width == p2.width


def test_var_from_def():
    mod = Generator("mod")
    v1 = mod.var("v1", 2, size=[2, 3])
    v2 = mod.var_from_def(v1, "v2")
    assert v2.width == v1.width
    assert v2.name == "v2"
    p1 = mod.input("p1", 42)
    v3 = mod.var_from_def(p1, "v3")
    assert v3.width == 42


def test_raw_import():
    mod = Generator("mod")
    mod.internal_generator.add_raw_import("pkg_name")
    src = verilog(mod, optimize_passthrough=False)["mod"]
    assert "module mod \n  import pkg_name::*;\n(\n);\n\nendmodule" in src


def test_generator_port_connected():
    child = Generator("child")
    parent = Generator("parent")
    in1 = parent.var("a", 1)
    in2 = child.input("a", 1)
    out1 = parent.output("b", 1)
    out2 = child.output("b", 1)
    un_c = child.input("c", 1)
    parent.add_child("inst", child, a=in1, b=out1)
    assert in2.connected()
    assert out2.connected()
    assert not un_c.connected()


def test_port_type():
    mod = Generator("mod")
    struct = PackedStruct("config", [("read", 16), ("write", 16)])
    # make it external to test if we can skip the generation
    struct.external = True
    in1 = mod.input("in1", struct)
    enum = kratos.enum("color", ["color", "green"])
    enum.external = True
    in2 = mod.input("in_enum", enum)
    # raw types
    param = mod.param("type_t", is_raw_type=True)
    param.value = "pkg::new_type_t"
    mod.parameter("type2", is_raw_type=True, initial_value="pkg::new_type_t")
    mod.parameter("type3", is_raw_type=True, initial_value="pkg::new_type_t")
    p_width = mod.parameter("P_width", value=5)
    a_in = mod.input("a", p_width)

    # raw type port
    # width 5 should be ignored. this is just for testing
    p = mod.input("in_raw", width=5)
    p.raw_type_param = param
    assert param.param_type == kratos.ParamType.RawType
    src = verilog(mod)["mod"]
    assert "type type_t" in src
    assert "input config in1" in src
    assert "input color in_enum" in src
    assert "input type_t in_raw" in src
    assert "parameter type type_t = pkg::new_type_t" in src

    # test raw type codegen with a parent
    parent = Generator("parent")
    p_raw = parent.parameter("p_type", is_raw_type=True,
                             value="pkg::old_value_t")
    param.raw_type_param = p_raw
    param.value = p_raw

    parent.port_from_def(in1)
    parent.port_from_def(in2)
    parent.port_from_def(p)
    parent.port_from_def(a_in, check_param=False)
    parent.add_child("inst", mod, in1=parent.ports.in1,
                     in_enum=parent.ports.in_enum,
                     in_raw=parent.ports.in_raw,
                     a=parent.ports.a)

    src = verilog(parent)["parent"]
    assert "type_t(p_type)" in src
    assert ".P_width(32'h5)" in src


def test_param_size(check_gold):
    mod = Generator("mod")
    width = mod.parameter("WIDTH", 16, 8)
    addr_width = mod.parameter("ADDR_WIDTH", 16, 8)
    data_in = mod.input("data_in", width)
    addr = mod.input("addr", addr_width)
    clk = mod.clock("clk")
    mem = mod.var("data", width, size=2 ** addr_width)

    @always_ff((posedge, clk))
    def code():
        mem[addr] = data_in

    mod.add_always(code)

    check_gold(mod, "test_param_size")


def test_param_packed_struct_array(check_gold):
    mod = Generator("mod")
    struct = PackedStruct("bus", [("read", 16), ("write", 16)])
    param = mod.parameter("P", 32, 42)
    out = mod.output("out", struct, size=[param])
    var = mod.var("v", struct, size=[param])
    mod.add_stmt(out.assign(var))

    check_gold(mod, "test_param_packed_struct_array")

    assert out[0]["write"].width == 16


def test_param_copy_def():
    enum = kratos.enum("enum_t", ["A", "B"])
    enum.external = True
    child = Generator("child")
    p1 = child.parameter("P1")
    p2 = child.parameter("P2", is_raw_type=True)
    p3 = child.parameter("P3", enum)
    top = Generator("top")
    top.param_from_def(p1)
    top.param_from_def(p2, "P_t")
    top.param_from_def(p3)
    src = verilog(top)["top"]

    assert "parameter P1" in src
    assert "parameter enum_t P3" in src
    assert "parameter type P_t" in src


def test_multi_gen():
    from kratos.util import enable_multi_generate
    enable_multi_generate()

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            a = self.input("a", 1)
            b = self.output("b", 1)
            self.wire(a, b)
    mod1 = Mod()
    verilog(mod1)
    mod2 = Mod()
    mod2.ports.a.name = "a2"
    src = verilog(mod2)
    assert "mod" not in src
    assert "mod_unq0" in src


def test_param_resize():
    mod = Generator("mod")
    p = mod.parameter("P")
    a = mod.var("a", 5)
    b = mod.var("b", 5)
    mod.add_stmt(b.assign(a + p))


def test_clog2_var():
    from kratos import clog2
    mod = Generator("mod")
    num = mod.parameter("num", value=32)
    num2 = mod.parameter("num2", value=1)
    c = clog2(num2)
    mod.input("a", 16, size=clog2(num))
    b = mod.input("b", 16, size=clog2(num), packed=True)
    c = mod.var("c", c)

    src = verilog(mod, filename="test.sv", remove_unused=False)["mod"]
    assert "input logic [15:0] a [($clog2 (num))-1:0]" in src
    assert "input logic [($clog2 (num))-1:0] [15:0] b" in src
    assert "logic [$clog2 (num2)-1:0] c" in src
    # test slicing
    d = mod.var("d", 1)
    e = b[d]
    assert str(e) == "b[d]"
    f = c[d]
    assert str(f) == "c[d]"


def test_ssa_transform(check_gold):
    mod = Generator("mod", debug=True)
    a = mod.var("a", 4)
    b = mod.var("b", 4)
    c = mod.output("c", 4)

    @always_comb
    def func():
        a = 1
        a = 2
        if a == 2:
            a = b + a
        if a == 3:
            b = 2
        else:
            if a == 4:
                b = 3
            else:
                b = 4
                # this is not a latch
                a = 5
        c = a

    mod.add_always(func, ssa_transform=True)
    check_gold(mod, "test_ssa_transform", ssa_transform=True)
    # check if the local variable mapping is fixed
    # assign a_5 = (a_3 == 4'h4) ? a_3: a_4;
    # which corresponds to a = 5
    # notice that a should be pointing to a = b + a, since it's the last
    # time a gets assigned
    stmt = mod.get_stmt_by_index(12)
    scope = stmt.scope_context
    is_var, a_mapping = scope["a"]
    assert is_var
    # this is assign a_2 = b + a_1;
    stmt = mod.get_stmt_by_index(5)
    assert str(a_mapping) == str(stmt.left)

    # test enable table extraction
    from _kratos.passes import compute_enable_condition
    enable_map = compute_enable_condition(mod.internal_generator)
    assert len(enable_map) > 5


def test_enable_condition_always_ff():
    mod = Generator("mod")
    a = mod.var("a", 4)
    b = mod.var("b", 1)
    clk = mod.clock("clk")

    @always_ff((posedge, clk))
    def logic():
        if b:
            a = 0
        else:
            a = 1

    mod.add_always(logic)
    from _kratos.passes import compute_enable_condition
    enable_map = compute_enable_condition(mod.internal_generator)
    assert len(enable_map) == 2


if __name__ == "__main__":
    from conftest import check_gold_fn
    test_ssa_transform(check_gold_fn)



from typing import List, Union


class InterfaceWrapper:
    """Interface wrapper"""
    def __init__(self, interface):
        self._i = interface

    def input(self, name: str, width: int, size: Union[int, List[int]] = 1):
        return self._i.input(name, width, size)

    def output(self, name: str, width: int, size: Union[int, List[int]] = 1):
        return self._i.output(name, width, size)

    def var(self, name: str, width: int, size: Union[int, List[int]] = 1):
        return self._i.var(name, width, size)

    def __getitem__(self, item):
        if self._i.has_modport(item):
            return self._i.get_modport_ref(item)
        elif item in self._i:
            return self._i[item]
        raise AttributeError

    def __getattr__(self, item):
        return self[item]

    @property
    def internal_interface(self):
        return self._i



from kratos import Generator, TestBench, initial, assert_, delay, Sequence


def tb_dut_setup():
    dut = Generator("mod")
    dut.wire(dut.output("out", 1), dut.input("in", 1))
    dut.wire(dut.var("val", 1), dut.ports["in"])

    tb = TestBench()

    tb.add_child("dut", dut)
    in_ = tb.var("in", 1)
    out_ = tb.var("out", 1)
    tb.wire(dut.ports["in"], in_)
    tb.wire(out_, dut.ports["out"])
    return dut, tb


def test_tb_codegen(check_gold):
    dut, tb = tb_dut_setup()

    @initial
    def code():
        tb.vars["in"] = 1
        assert_(tb.vars.out == 1)
        # access internal signal
        assert_(dut.vars.val == 1)

    tb.add_always(code)

    src = tb.codegen()
    check_gold(src, "test_tb_codegen")


def test_tb_delay(check_gold):
    dut, tb = tb_dut_setup()

    @initial
    def code():
        delay(1, tb.vars["in"].assign(1))

    tb.add_always(code)
    src = tb.codegen()
    check_gold(src, "test_tb_delay")


def test_tb_sequence(check_gold):
    from kratos.util import clock
    from kratos import PropertyAction

    dut, tb = tb_dut_setup()
    # add a clock and wire them together
    tb.wire(dut.clock("clk"), clock(tb.var("clk", 1)))

    seq = Sequence(tb.vars["in"] == 1)
    seq.imply(tb.vars.out == 1).wait(1).imply(tb.vars.out == 0)

    p = tb.property("test_out", seq)
    p.action = PropertyAction.Assert

    src = tb.codegen()
    check_gold(src, "test_tb_sequence")


if __name__ == "__main__":
    from conftest import check_gold_fn
    test_tb_sequence(check_gold_fn)



import _kratos
from .generator import Generator, transform_stmt_block, CodeBlockType, \
    InitialCodeBlock, VarProxy
from _kratos import get_fn_ln


def assert_(expr):
    if isinstance(expr, _kratos.Var):
        stmt = _kratos.AssertValueStmt(expr)
    else:
        assert isinstance(expr, _kratos.Property)
        expr.action = _kratos.PropertyAction.Assert
        stmt = _kratos.AssertPropertyStmt(expr)
    if expr.generator.debug:
        stmt.add_fn_ln(get_fn_ln())
    return stmt


def assume(prop):
    prop.action = _kratos.PropertyAction.Assume
    stmt = _kratos.AssertPropertyStmt(prop)
    return stmt


def cover(prop):
    prop.action = _kratos.PropertyAction.Cover
    stmt = _kratos.AssertPropertyStmt(prop)
    return stmt


def delay(num, stmt):
    assert isinstance(stmt, _kratos.AssignStmt)
    stmt.delay = num
    return stmt


class TestBench:
    # disable pytest collection
    __test__ = False

    def __init__(self, top_name: str = "TOP"):
        self.__tb = _kratos.TestBench(Generator.get_context(), top_name)
        self.__child_generator = {}
        self.debug = False
        self.internal_generator = self.__tb.top()

        # proxy
        self.vars = VarProxy(self)

    def var(self, name, width, size: int = 1, is_signed: bool = False):
        return self.__tb.var(name, width, size, is_signed)

    def add_child_generator(self, instance_name: str, generator: Generator):
        assert instance_name not in self.__child_generator
        self.__child_generator[instance_name] = generator
        self.__tb.add_child_generator(instance_name,
                                      generator.internal_generator)

    add_child = add_child_generator

    def codegen(self):
        return self.__tb.codegen()

    def add_stmt(self, stmt):
        if not isinstance(stmt, _kratos.Stmt):
            self.__tb.add_stmt(stmt.stmt())
        else:
            self.__tb.add_stmt(stmt)

    def wire(self, var1, var2):
        self.__tb.wire(var1, var2)

    def initial(self):
        return self.__tb.initial()

    def add_always(self, fn, comment=""):
        block_type, raw_sensitives, stmts = transform_stmt_block(self, fn)
        if block_type != CodeBlockType.Initial:
            raise NotImplementedError(
                "Only initial supported in test bench")
        if block_type == CodeBlockType.Combinational:
            node = None
            pass
        elif block_type == CodeBlockType.Initial:
            # it's a initial block
            init = InitialCodeBlock(self)
            for stmt in stmts:
                init.add_stmt(stmt)
            node = init
        else:
            node = None
        if comment:
            node.comment = comment

    add_code = add_always

    def property(self, property_name: str, sequence):
        return self.__tb.property(property_name, sequence)



from _kratos import Simulator as _Simulator
from .generator import Generator, PortType


# Python wrapper for the simulator
class Simulator:
    def __init__(self, generator: Generator):
        self._sim = _Simulator(generator.internal_generator)
        # get the clock and reset
        clks = generator.internal_generator.get_ports(PortType.Clock.value)
        if len(clks) == 1:
            self._clk = generator.ports[clks[0]]
        else:
            self._clk = None
        resets = generator.internal_generator.get_ports(PortType.AsyncReset.value)
        if len(resets) == 1:
            self._reset = generator.ports[resets[0]]
        else:
            self._reset = None

    def set(self, var, value):
        self._sim.set(var, value)

    def get(self, var):
        if len(var.size) > 1 or var.size[0] > 1:
            return self._sim.get_array(var)
        else:
            return self._sim.get(var)

    def cycle(self, n=1):
        if self._clk is None:
            raise RuntimeError("Single clock not found")
        for _ in range(n):
            self.set(self._clk, 1)
            self.set(self._clk, 0)

    def reset(self, reset_high=True):
        if self._reset is None:
            raise RuntimeError("Single reset not found")
        if reset_high:
            self.set(self._reset, 1)
            self.set(self._reset, 0)
        else:
            self.set(self._reset, 0)
            self.set(self._reset, 1)



from _kratos.formal import remove_async_reset as _remove_async_reset
from .passes import verilog
import tempfile
import os
import shutil
import subprocess


def output_btor(generator, filename, remove_async_reset=True, yosys_path="",
                quite=False, **kargs):
    # check yosys
    if yosys_path == "" or not os.path.isfile(yosys_path):
        # need to figure out yosys by ourselves
        yosys_path = shutil.which("yosys")
        assert yosys_path, "Yosys not found"
    # remove async reset if necessary
    if remove_async_reset:
        _remove_async_reset(generator.internal_generator)
    # abs path
    filename = os.path.abspath(filename)

    # I think most of the kratos semantics are supported by yosys
    # if not, you can also turn on the flag to use sv2v to convert
    with tempfile.TemporaryDirectory() as temp:
        verilog_filename = os.path.join(temp, "verilog.sv")
        verilog(generator, filename=verilog_filename, **kargs)
        # need to output the yosys file
        yosys_file = os.path.join(temp, "convert.ys")
        with open(yosys_file, "w+") as f:
            f.write("read_verilog -formal -sv {0}\n".format(verilog_filename))
            f.write("prep -top {0}\n".format(generator.name))
            f.write("flatten\n")
            f.write("write_btor {0}\n".format(filename))

        # call yosys
        extra_args = ["-s"]
        if quite:
            extra_args = ["-q"] + extra_args
        subprocess.check_call([yosys_path] + extra_args + [yosys_file])



from kratos import Generator, always_ff, verilog, posedge, always_comb
from kratos.util import reduce_add
import _kratos
import sqlite3
import tempfile
import os


def test_db_dump():
    mod = Generator("mod", True)
    comb = mod.combinational()
    comb.add_stmt(mod.var("a", 1).assign(mod.var("b", 1)))

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        # hashing and generate verilog
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * from breakpoint")
        result = c.fetchall()
        assert len(result) == 1
        conn.close()


def test_debug_mock():
    # this is used for the runtime debugging
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)

            # ports
            self.in1 = self.input("in1", 16)
            self.in2 = self.input("in2", 16)
            self.out = self.output("out", 16)

            self.add_always(self.code)

        @always_comb
        def code(self):
            if self.in1 == 2:
                self.out = 2
            elif self.in1 == 1:
                self.out = 0
            elif self.in2 == 1:
                self.out = 1
            else:
                self.out = 3

    with tempfile.TemporaryDirectory() as temp:
        mod = Mod()
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                insert_verilator_info=True, debug_db_filename=debug_db)
        # make sure the verilator public is there
        with open(filename) as f:
            content = f.read()
        assert "verilator public" in content


def test_seq_debug():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            # ports
            self.in_ = self.input("in1", 1)
            self.clock("clk")
            for i in range(4):
                self.output("out{0}".format(i), 1)

            self.add_always(self.code1)
            self.add_always(self.code2)

        @always_comb
        def code1(self):
            if self.in_ == 0:
                self.ports.out0 = 0
                self.ports.out1 = 0
            else:
                self.ports.out0 = 1
                self.ports.out1 = 1

        @always_ff((posedge, "clk"))
        def code2(self):
            if self.in_ == 0:
                self.ports.out2 = 0
                self.ports.out3 = 0
            else:
                self.ports.out2 = 1
                self.ports.out3 = 1

    mod = Mod()
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM breakpoint WHERE id=7")
        result = c.fetchall()
        assert len(result) == 1
        conn.close()


def test_context():
    class Mod(Generator):
        def __init__(self, width):
            super().__init__("mod", True)
            in_ = self.input("in", width)
            out = self.output("out", width)
            sel = self.input("sel", width)
            # test self variables
            self.out = out
            self.width = width

            @always_comb
            def code():
                if sel:
                    out = 0
                else:
                    for i in range(width):
                        out[i] = 1
            self.add_always(code)

    mod = Mod(4)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM context_variable")
        variables = c.fetchall()
        assert len(variables) > 20
        conn.close()


def test_clock_interaction():
    mods = []
    num_child = 4
    for i in range(num_child):
        mod = Generator("mod", True)
        in_ = mod.input("in", 4)
        out_ = mod.output("out", 4)
        clk = mod.clock("clk")
        seq = mod.sequential((posedge, clk))
        seq.add_stmt(out_.assign(in_))
        mods.append(mod)
    parent = Generator("parent", True)
    clk = parent.clock("clk")
    in_ = parent.input("in", 4)
    out = parent.output("out", 4)
    for i, mod in enumerate(mods):
        parent.add_child("mod{0}".format(i), mod)
        parent.wire(mod.ports.clk, clk)
        if i == 0:
            continue
        parent.wire(mod.ports["in"], mods[i - 1].ports.out)
    parent.wire(mods[0].ports["in"], in_)
    parent.wire(out, mods[-1].ports.out)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(parent, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)


def test_assert():
    from kratos import assert_
    mod = Generator("mod", True)
    in_ = mod.input("in", 1)
    out_ = mod.output("out", 1)

    @always_comb
    def code():
        # we introduce this bug on purpose
        out_ = in_ - 1
        assert_(out_ == in_)

    mod.add_always(code)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        with open(filename) as f:
            content = f.read()
            assert "assert (out == in) else" in content
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM breakpoint")
        lines = c.fetchall()
        assert len(lines) == 2
        # they are only one line apart
        assert abs(lines[0][3] - lines[1][3]) == 1
        conn.close()
    # once we remove the assertion, it should not be there
    _kratos.passes.remove_assertion(mod.internal_generator)
    src = verilog(mod)[0]["mod"]
    assert "assert" not in src


def test_wire():
    mod = Generator("mod", True)
    in_ = mod.input("in", 1)
    out_ = mod.output("out", 1)
    # context
    a = 1
    mod.wire(out_, in_)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM breakpoint")
        assert len(c.fetchall()) == 1
        c.execute("""SELECT variable.value FROM variable, context_variable
                     WHERE context_variable.name = ? AND context_variable.variable_id = variable.id""", "a")
        v = int(c.fetchone()[0])
        assert v == a
        conn.close()


def test_inst_id():
    def create_mod():
        m = Generator("mod", True)
        in_ = m.input("in", 1)
        out = m.output("out", 1)
        comb = m.combinational()
        comb.add_stmt(out.assign(in_))
        return m
    mod = Generator("parent", True)
    input_ = mod.input("in", 1)
    output = mod.output("out", 1)
    mods = [create_mod() for _ in range(2)]
    expr = None
    for i, m_ in enumerate(mods):
        mod.add_child("mod{0}".format(i), m_)
        mod.wire(input_, m_.ports["in"])
        if expr is None:
            expr = m_.ports["out"]
        else:
            expr = expr & m_.ports["out"]
    mod.wire(output, expr)

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db,
                optimize_passthrough=False)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM instance")
        res = c.fetchall()
        assert len(res) == 3
        conn.close()


def test_empty():
    from kratos.debug import dump_external_database
    mod = Generator("mod", True)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        dump_external_database([mod], "mod", debug_db)


def test_nested_scope():
    from kratos import clog2
    mod = Generator("FindHighestBit", True)
    width = 4
    data = mod.input("data", width)
    h_bit = mod.output("h_bit", clog2(width))
    done = mod.var("done", 1)

    @always_comb
    def find_bit():
        done = 0
        h_bit = 0
        for i in range(width):
            if ~done:
                if data[i]:
                    done = 1
                    h_bit = i
    mod.add_always(find_bit, label="block")
    verilog(mod, insert_debug_info=True)
    block = mod.get_marked_stmt("block")
    last_if = block[-1]
    for i in range(len(last_if.then_[-1].then_)):
        stmt = last_if.then_[-1].then_[i]
        context = stmt.scope_context
        if len(context) > 0:
            assert "i" in context
            is_var, var = context["i"]
            assert not is_var
            assert var == "3"


def test_array_packed():
    from kratos import PackedStruct
    mod = Generator("mod", True)
    aa = mod.var("a", 2, size=(2, 4), packed=True)
    struct = PackedStruct("s", [("read", 16, False),
                                ("data", 16, False)])
    ss = mod.var_packed("s", struct)
    mod.add_stmt(aa.assign(4))
    setattr(mod, "sss", ss)
    mod.add_stmt(ss["read"].assign(0))
    mod.add_stmt(ss["data"].assign(1))

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, debug_db_filename=debug_db, insert_debug_info=True)
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT variable.value, generator_variable.name FROM variable, generator_variable WHERE variable.id = generator_variable.variable_id")
        vars_ = c.fetchall()
        c.execute("SELECT variable.value, context_variable.name FROM variable, context_variable WHERE variable.id = context_variable.variable_id")
        vars_ += c.fetchall()
        correct_struct, correct_array, correct_self = False, False, False
        for value, name in vars_:
            if value == "a[1][3]" and name == "aa.1.3":
                correct_array = True
            if value == "s.read" and name == "ss.read":
                correct_struct = True
            if "self.sss" in name:
                correct_self = True
        assert correct_array and correct_struct, correct_self
        conn.close()


def test_multiple_instance():
    class Mod(Generator):
        def __init__(self, width, param):
            super().__init__("child_{0}_{1}".format(width, param), True)
            in_ = self.input("child_in", width)
            out_ = self.output("child_out", width)

            @always_comb
            def code():
                out_ = 0
                for i in range(param):
                    out_ = in_ + param

            self.add_always(code)

    top = Generator("parent", True)
    top_in = top.input("in", 8)
    top_out = top.output("out", 8)
    children = []
    out_ports = []
    num_instance = 3
    for i in range(num_instance):
        child = Mod(8, i + 2)
        children.append(child)
    for i in range(num_instance):
        child = children[i]
        top.add_child("child_{0}".format(i), child,
                      child_in=top_in)
        out_ports.append(child.ports.child_out)

    top.wire(top_out, reduce_add(*out_ports))

    # testing code
    with tempfile.TemporaryDirectory() as temp:
        filename = os.path.join(temp, "test.sv")
        verilog(top, filename=filename, insert_debug_info=True, debug_fn_ln=True)
        # check the id and line info
        with open(filename) as f:
            content = f.read()
            assert ".KRATOS_INSTANCE_ID(32'h2))" in content
        # make sure that the line is mapped correctly
        with open(__file__) as f:
            lines = f.readlines()
            index = -1
            for i in range(len(lines)):
                if "out_ = in_ + param" in lines[i]:
                    index = i
                    break
            assert index != -1
        line_no = index + 1  # 1 offset for line number
        with open(filename + ".debug") as f:
            import json
            content = json.load(f)
        count = 0
        for _, entry in content.items():
            __, ln = entry[0]
            if ln == line_no:
                count += 1
        assert count == num_instance * (num_instance + 3) / 2


def get_line(line):
    with open(__file__) as f:
        lines = f.readlines()
        lines = [l.rstrip() for l in lines]
    return lines.index(line) + 1


def test_ssa_debug():
    mod = Generator("mod", debug=True)
    a = mod.var("a", 4)
    clk = mod.clock("clk")
    b = mod.var("b", 1)
    loop_size = 4

    @always_comb
    def logic1():
        a = 0
        if b:
            for i in range(loop_size):
                a = a + i

    @always_ff((posedge, clk))
    def logic2():
        if a == 1:
            b = 1
        else:
            b = 0

    mod.add_always(logic1, ssa_transform=True)
    mod.add_always(logic2)

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db, ssa_transform=True)
        # assert the line number tracking
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        idx = get_line("                a = a + i")
        c.execute("SELECT * FROM breakpoint WHERE line_num=?", (idx,))
        result = c.fetchall()
        assert len(result) == loop_size
        assert "a_4" in result[0][-1]
        # check the context variable
        c.execute("SELECT * FROM context_variable WHERE context_variable.name = 'i'")
        result = c.fetchall()
        assert len(result) == loop_size
        conn.close()


if __name__ == "__main__":
    test_ssa_debug()



from kratos import Interface, Generator, always_ff, posedge, verilog
import tempfile
import os


class ConfigInterface(Interface):
    def __init__(self):
        Interface.__init__(self, "Config")
        width = 8
        # local variables
        read = self.var("read_data", width)
        write = self.var("write_data", width)
        r_en = self.var("r_en", 1)
        w_en = self.var("w_en", 1)
        # common ports
        clk = self.clock("clk")

        # define master -> slave ports
        # and slave -> master ports
        m_to_s = [write, r_en, w_en]
        s_to_m = [read]

        # define master and slave
        self.master = self.modport("Master")
        self.slave = self.modport("Slave")
        for port in m_to_s:
            self.master.set_output(port)
            self.slave.set_input(port)
        for port in s_to_m:
            self.master.set_input(port)
            self.slave.set_output(port)

        # both of them need clock
        self.master.set_input(clk)
        self.slave.set_input(clk)


def test_modport_io(check_gold):
    config = ConfigInterface()

    class Master(Generator):
        def __init__(self):
            Generator.__init__(self, "Master")

            # instantiate the interface
            self.bus = self.interface(config.master, "bus", is_port=True)

            # some logic to loop the read and write
            # cycle
            self.counter = self.var("counter", 8)

            # we wire counter value to the write data
            self.wire(self.bus.write_data, self.counter)

            # always_ff on the posedge of clock from the interface
            @always_ff((posedge, self.bus.clk))
            def logic():
                if self.counter % 4 == 0:
                    self.bus.r_en = 1
                    self.bus.w_en = 0
                elif self.counter % 4 == 1:
                    self.bus.r_en = 0
                    self.bus.w_en = 1
                else:
                    self.bus.r_en = 0
                    self.bus.w_en = 0

            @always_ff((posedge, self.bus.clk))
            def update():
                self.counter = self.counter + 1

            self.add_always(logic)
            self.add_always(update)

    class Slave(Generator):
        def __init__(self):
            Generator.__init__(self, "Slave")

            # instantiate the interface
            self.bus = self.interface(config.slave, "bus", is_port=True)

            self.value = self.var("value", 8)

            # just read and write out
            @always_ff((posedge, self.bus.clk))
            def logic():
                if self.bus.r_en:
                    self.value = self.bus.write_data
                elif self.bus.w_en:
                    self.bus.read_data = self.value

            self.add_always(logic)

    class Top(Generator):
        def __init__(self):
            Generator.__init__(self, "Top")

            # instantiate master and slave
            self.master = Master()
            self.slave = Slave()
            self.add_child("master", self.master)
            self.add_child("slave", self.slave)

            # clock will be from outside
            clk = self.clock("clk")

            # instantiate the interface bus
            # notice that we're using config, not the modport
            # version such as config.master
            self.bus = self.interface(config, "bus_top")

            # just need to wire things up
            self.wire(self.bus.clk, clk)
            self.wire(self.master.bus, self.bus)
            self.wire(self.slave.bus, self.bus)
            # the following also works
            # self.wire(self.master.bus, bus.Master)
            # self.wire(self.slave.bus, bus.Slave)

    top = Top()
    check_gold(top, "test_modport_io")
    assert str(top.bus.read_data) == "bus_top.read_data"


def test_port_interface():
    mod = Generator("mod")
    mod.interface(ConfigInterface(), "port_interface", is_port=True)

    with tempfile.TemporaryDirectory() as temp:
        filename = os.path.join(temp, "mod.sv")
        verilog(mod, filename=filename)
        with open(filename) as f:
            content = f.read()
            assert "endinterface" in content


if __name__ == "__main__":
    test_port_interface()




import sys
from typing import Union, List
import os
import math
from _kratos import mux as _mux, comment as _comment, \
    create_stub as _create_stub
import _kratos
import enum
import functools
import operator


class CLIColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_src(src, line_no: Union[List[int], int], offset: int = 1,
              code_range: int = 2):
    if os.path.isfile(src):
        with open(src) as f:
            lines = f.readlines()
    else:
        lines = src.split("\n")
    line_start = 0
    line_end = len(lines) - 1
    if isinstance(line_no, int):
        line_no = [line_no]
    for line in line_no:
        line_start = max(0, line - offset - code_range)
        line_end = min(len(lines) - 1, line - offset + code_range)
    # print a line
    print(CLIColors.OKBLUE + "-" * 80 + CLIColors.ENDC, file=sys.stderr)
    for idx in range(line_start, line_end + 1):
        if idx + offset in line_no:
            print(CLIColors.FAIL + ">", lines[idx] + CLIColors.ENDC,
                  file=sys.stderr, end="")
        else:
            print(CLIColors.OKGREEN + " ", lines[idx] + CLIColors.ENDC,
                  file=sys.stderr, end="")

    # print a line
    print(CLIColors.OKBLUE + "-" * 80 + CLIColors.ENDC, file=sys.stderr)


def clog2(x: Union[int, _kratos.Var]) -> Union[int, _kratos.Var]:
    if isinstance(x, _kratos.Var):
        from kratos.func import get_built_in
        fn = get_built_in(x.generator, "clog2")
        return fn(x)
    else:
        if x == 0:
            return 0
        return int(math.ceil(math.log2(x)))


def flog2(x: int) -> int:
    if x == 0:
        return 0
    return int(math.floor(math.log2(x)))


# these are helper functions tp construct complex expressions
def __reduce(func, args):
    r = functools.reduce(func, args)
    if isinstance(r, (list, tuple)) and len(r) == 1:
        return r[0]
    return r


def reduce_or(*args):
    return __reduce(operator.or_, args)


def reduce_and(*args):
    return __reduce(operator.and_, args)


def reduce_add(*args):
    return __reduce(operator.add, args)


def reduce_mul(*args):
    return __reduce(operator.mul, args)


def concat(*args):
    def _concat(a, b):
        return a.concat(b)
    return __reduce(_concat, args)


def ext(var, target_width):
    return var.extend(target_width)


def mux(cond, left, right):
    return _mux(cond, left, right)


class VarCastType(enum.Enum):
    Signed = _kratos.VarCastType.Signed
    Unsigned = _kratos.VarCastType.Unsigned
    Clock = _kratos.VarCastType.Clock
    AsyncReset = _kratos.VarCastType.AsyncReset
    Enum = _kratos.VarCastType.Enum
    Resize = _kratos.VarCastType.Resize


def cast(var, cast_type, **kargs):
    assert isinstance(var, _kratos.Var)
    _v = var.cast(cast_type.value)
    for k, v in kargs.items():
        setattr(_v, k, v)
    return _v


def signed(var):
    return cast(var, VarCastType.Signed)


def resize(var, target_width):
    return cast(var, VarCastType.Resize, target_width=target_width)


def unsigned(var):
    return cast(var, VarCastType.Unsigned)


def clock(var):
    return cast(var, VarCastType.Clock)


def async_reset(var):
    return cast(var, VarCastType.AsyncReset)


def const(value: int, width: int, is_signed: bool = False):
    return _kratos.constant(value, width, is_signed)


def comment(comment_str):
    return _comment(comment_str)


def create_stub(generator, filename=""):
    s = _create_stub(generator.internal_generator)
    if filename:
        with open(filename) as f:
            f.write(s)
    return s


def max_value(values):
    s = set()
    for x in values.values():
        s.add(x)
    return max(s)


def enable_multi_generate():
    # need to enable if verilog() function is called multiple times
    # and all of them will go to the same RTL code
    from kratos import Generator
    Generator.get_context().track_generated = True


# bit vector style syntax
class ConstConstructor:
    def __getitem__(self, width):
        class ConstWidth:
            def __call__(self, value, is_signed=False):
                return _kratos.constant(value, width, is_signed)

        return ConstWidth()


Const = ConstConstructor()

# also create an alias
ternary = mux


# helper function to interface with magma
def to_magma(kratos_inst, flatten_array=False, top_name=None, **kargs):  # pragma: no cover
    import magma as m
    from kratos import verilog, Generator
    from _kratos import create_wrapper_flatten
    if flatten_array:
        inst = create_wrapper_flatten(kratos_inst.internal_generator,
                                      kratos_inst.name + "_W")
        inst = Generator(kratos_inst.name,
                         internal_generator=inst)
        # only add the child, nothing more
        inst.add_child(kratos_inst.instance_name, kratos_inst, python_only=True)
        kratos_inst = inst
        if top_name is not None:
            kratos_inst.instance_name = top_name
    circ_name = kratos_inst.name
    internal_gen = kratos_inst.internal_generator
    ports = internal_gen.get_port_names()
    io = {}
    for port_name in ports:
        port = kratos_inst.ports[port_name]
        width = port.width
        size = port.size
        dir_ = m.In if port.port_direction == _kratos.PortDirection.In \
            else m.Out
        type_ = port.port_type
        signed = port.signed
        if type_ == _kratos.PortType.Clock:
            type_value = m.Clock
        elif type_ == _kratos.PortType.AsyncReset:
            if port.active_high or port.active_high is None:
                type_value = m.AsyncReset
            else:
                type_value = m.AsyncResetN
        else:
            # normal logic/wire, loop through the ndarray to construct
            # magma arrays, notice it's in reversed order
            type_value = m.Bits[width] if not signed else m.SInt[width]
            if len(size) > 1 or size[0] > 1:
                for idx, array_width in enumerate(reversed(size)):
                    type_value = m.Array[array_width, type_value]
        io[port_name] = (dir_(type_value))
    magma_io = m.IO(**io)

    class _Definition(m.Circuit):
        name = circ_name
        io = magma_io
        kratos = kratos_inst
    os.makedirs(".magma", exist_ok=True)
    filename = f".magma/{circ_name}-kratos.sv"
    enable_multi_generate()
    # multiple definition inside the kratos instance is taken care of by
    # the track definition flag
    verilog(kratos_inst, filename=filename, track_generated_definition=True,
            **kargs)
    with open(filename, 'r') as f:
        _Definition.verilogFile = f.read()

    return _Definition



from kratos import Simulator, Generator, posedge, always_ff, negedge


def test_sim_reg():
    class AsyncReg(Generator):
        def __init__(self, width, debug=False):
            super().__init__("Register", debug)

            self.in_ = self.input("in", width)
            self.out = self.output("out", width)
            self.clk = self.clock("clk")
            self.rst = self.reset("rst")

            self.add_always(self.seq_code_block)

        @always_ff((posedge, "clk"), (negedge, "rst"))
        def seq_code_block(self):
            if ~self.rst:
                self.out = 0
            else:
                self.out = self.in_

    reg = AsyncReg(16)
    sim = Simulator(reg)
    val = sim.get(reg.out)
    assert val is None
    # reset
    sim.reset(False)
    val = sim.get(reg.out)
    assert val == 0
    # put in some values
    for v in range(1, 10 + 1):
        sim.set(reg.in_, v)
        val = sim.get(reg.out)
        # 1 cycle delay
        assert val == v - 1
        # toggle the clock
        sim.cycle()
        val = sim.get(reg.out)
        assert val == v


def test_expr():
    mod = Generator("mod")
    a = mod.input("a", 16)
    b = mod.output("b", 16)
    e = a + a
    for i in range(4):
        e = e + a
    mod.add_stmt(b.assign(e))

    sim = Simulator(mod)
    assert sim.get(b) is None
    sim.set(a, 2)
    assert sim.get(b) == 12


if __name__ == "__main__":
    test_expr()



import _kratos
from .generator import Generator
from typing import List


def extract_symbol_table(generator: Generator):
    # this has to be run after the unification pass
    # if you are using kratos from a different system
    # this function serves a reference implementation of how to obtain
    # data from the python side
    from queue import Queue
    var_table = {}
    str_table = {}
    gen_queue = Queue()
    gen_queue.put(generator)
    while not gen_queue.empty():
        gen: Generator = gen_queue.get()
        if gen.debug:
            # introspect the variable tables
            var_entry = {}
            str_entry = {}
            variables = vars(gen)
            for name, var in variables.items():
                # since this is python, we use self to refer to itself
                name = "self." + name
                if isinstance(var, _kratos.Var):
                    # I think bundle -> packed struct will not work here
                    var_entry[name] = var
                elif isinstance(var, (bool, int, str)):
                    str_entry[name] = str(var)
            var_table[gen] = var_entry
            str_table[gen] = str_entry
        # push all the child generator to the queue
        children = gen.child_generator()
        for _, child in children.items():
            if child.internal_generator.parent_generator() is not None:
                # it could be removed
                gen_queue.put(child)
    return var_table, str_table


def enable_runtime_debug(generator: Generator):
    # insert breakpoints
    _kratos.inject_instance_ids(generator.internal_generator)
    _kratos.inject_debug_break_points(generator.internal_generator)


def dump_debug_database(generator: Generator, filename: str):
    db = _kratos.DebugDataBase()
    raw_symbol_table, raw_str_table = extract_symbol_table(generator)
    db.set_break_points(generator.internal_generator)
    # convert raw table to internal generator
    symbol_table = {}
    str_table = {}
    for gen, value in raw_symbol_table.items():
        symbol_table[gen.internal_generator] = value
    for gen, value in raw_str_table.items():
        str_table[gen.internal_generator] = value
    db.set_variable_mapping(symbol_table)
    # store the string values
    db.set_variable_mapping(str_table)
    # insert other metadata information
    db.set_stmt_context(generator.internal_generator)
    db.save_database(filename)


def dump_external_database(generators: List[Generator], top_name: str, filename: str):
    # create a dummy generator at the top
    top = Generator(top_name, True)
    for gen in generators:
        # remove top name with given one
        instance_name = gen.instance_name.split(".")[1:]
        if instance_name:
            top.add_child(".".join(instance_name), gen)
        else:
            # only has one generator
            assert len(generators) == 1
            gen.instance_name = top_name
            top = gen
    dump_debug_database(top, filename)



from kratos import Generator, verilog, const, always_comb
import tempfile
import os
import filecmp


class PassThroughMod(Generator):
    def __init__(self, is_clone: bool = False):
        super().__init__("mod1", True, is_clone)
        self.in_ = self.input("in", 1)
        self.out_ = self.output("out", 1)
        self.wire(self.out_, self.in_)


def test_regression_1(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("Top")
            self.in_ = self.input("i_data_in", 2)
            self.out_ = self.output("o_data_out", 2)

            child1 = PassThroughMod()
            self.add_child("child1", child1)
            self.wire(child1.in_, self.in_[0])
            self.wire(self.out_[0], child1.out_)

            child2 = PassThroughMod()
            self.add_child("child2", child2)
            self.wire(self.in_[1], child2.in_)
            self.wire(self.out_[1], child2.out_)

    mod = Mod()
    check_gold(mod, "test_regression_1", optimize_passthrough=False)


def test_regression_2():
    from kratos.func import dpi_function

    @dpi_function(8)
    def func(arg0):
        pass

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 8)
            self._reg = self.var("val", 8)

            self.add_always(self.code)

        @always_comb
        def code(self):
            # because the types are inferred
            # implicit const conversion doesn't work here
            self._reg = func(const(1, 2))
            self._out = func(const(1, 2))

    mod = Mod()
    mod_src = verilog(mod)
    assert "func" in mod_src[1]


def test_regression_concat():
    from kratos import concat
    parent = Generator("parent")
    for i in range(2):
        child = Generator("child")
        in_ = child.input("in", 1)
        out_ = child.output("out", 1)
        child.wire(out_, in_)
        parent.add_child("child{0}".format(i), child)
    in_ = parent.input("in", 1)
    out_2 = parent.output("out2", 2)
    for i in range(2):
        parent.wire(in_, parent["child{0}".format(i)].ports["in"])
    parent.wire(out_2,
                concat(parent["child0"].ports.out, parent["child1"].ports.out))
    src = verilog(parent, optimize_passthrough=False)["parent"]
    assert "assign out2 = {child0_out, child1_out};" in src


def test_regression_flatten_array_param():
    from _kratos import create_wrapper_flatten
    mod = Generator("mod")
    p = mod.parameter("param", value=16)
    mod.input("in", width=p)
    inst = create_wrapper_flatten(mod.internal_generator, "wrapper")
    gen = Generator("", internal_generator=inst)
    src = verilog(gen)["wrapper"]
    assert "parameter param = 32'h10" in src


if __name__ == "__main__":
    test_regression_flatten_array_param()



from kratos import IRVisitor, Generator, PortDirection, Port, VarSlice, \
    VarVarSlice, Attribute


def test_attribute():
    mod = Generator("mod")
    a = mod.input("a", 1)
    b = mod.output("b", 1)
    mod.wire(b, a)
    stmt = mod.get_stmt_by_index(0)

    class TestAttribute(Attribute):
        def __init__(self):
            Attribute.__init__(self)
            self.value = 42
            self.value_str = "42"

    stmt.add_attribute(TestAttribute())

    assert len(mod.get_stmt_by_index(0).get_attributes()) > 0
    attr = mod.get_stmt_by_index(0).get_attributes()[0].get()
    assert attr.value == 42
    assert attr.value_str == "42"
    assert attr.type_str == "python"


def test_ir_port(check_gold):
    def change_name(generator):
        class Visitor(IRVisitor):
            def __init__(self):
                IRVisitor.__init__(self)

            def visit(self, node):
                if isinstance(node, Port):
                    # rename the output port
                    if node.name == "out":
                        node.name = "test"

        visitor = Visitor()
        visitor.visit_root(generator)

    class Mod1(Generator):
        def __init__(self):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 1, PortDirection.Out)
            self.wire(self.out_, self.in_)

    mod = Mod1()
    check_gold(mod, "test_ir_port",
               additional_passes={"name_change": change_name})


def test_ir_slice():
    def collect_slice(generator):
        class Visitor(IRVisitor):
            def __init__(self):
                IRVisitor.__init__(self)
                self.hit_slice = False
                self.hit_var_slice = False

            def visit(self, node):
                if isinstance(node, VarVarSlice):
                    assert node.sliced_by_var
                    self.hit_var_slice = True
                elif isinstance(node, VarSlice):
                    assert not node.sliced_by_var
                    self.hit_slice = True

        visitor = Visitor()
        visitor.visit_root(generator)
        return visitor

    class Mod(Generator):
        def __init__(self):
            Generator.__init__(self, "mod")
            a = self.var("a", 8)
            b = self.var("b", 3)
            d = a[2]
            c = a[b]
            self.wire(self.output("out0", 1), d)
            self.wire(self.output("out1", 1), c)

    mod = Mod()
    v = collect_slice(mod.internal_generator)
    assert v.hit_slice
    assert v.hit_var_slice


def test_ir_annotation_wiring():
    from kratos import Attribute, verilog
    import kratos
    parent = Generator("parent")
    child = Generator("child")
    parent.add_child("child_inst", child)
    port = child.input("target_port", 1)
    child.wire(child.output("out", 1), port)
    tag = "target"
    port.add_attribute(Attribute.create(tag))

    # this is a pass
    def wire_up_port(generator):
        class WireUpVisitor(IRVisitor):
            def __init__(self):
                IRVisitor.__init__(self)
                self.tag = tag

            def visit(self, node):
                if isinstance(node, kratos.Port) and node.has_attribute(self.tag):
                    # need to wire it to the instances parent and up
                    gen = node.generator
                    parent_gen = gen.parent_generator()
                    child_port = node
                    while parent_gen is not None:
                        # create a port based on the target's definition
                        p = parent_gen.port(child_port)
                        parent_gen.wire(child_port, p)
                        # move up the hierarchy
                        child_port = p
                        parent_gen = parent_gen.parent_generator()

        v = WireUpVisitor()
        v.visit_root(generator)

    wire_up_port(parent.internal_generator)

    assert "target_port" in parent.ports
    # check all the connectivity
    verilog(parent)


if __name__ == "__main__":
    test_ir_annotation_wiring()



import subprocess
import os
import pytest
import tempfile


# get all example code
root_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
example_dir = os.path.join(root_dir, "examples")
files = os.listdir(example_dir)
filenames = []
for fn in files:
    fn_ = os.path.join(example_dir, fn)
    if os.path.isfile(fn_) and os.path.splitext(fn_)[-1] == ".py":
        # we only put the basename of filename here so that in the pytest
        # verbose printout, we won't have user's dir info
        filenames.append(fn)


@pytest.mark.parametrize("filename", filenames)
def test_examples(filename):
    filename = os.path.join(example_dir, filename)
    # use a temp dir
    with tempfile.TemporaryDirectory() as temp:
        args = ["python", filename]
        subprocess.check_call(args, cwd=temp)



from _kratos import PortBundleDefinition, PortType, PortDirection, \
    get_fn_ln


# a helper class to deal with port interface
class PortBundle:
    def __init__(self, debug=False):
        self.definition = PortBundleDefinition(self.__class__.__name__)
        self.debug = debug

    def input(self, name, width, is_signed=False, size=1,
              port_type=PortType.Data):
        self.definition.add_definition(name, width, size, is_signed,
                                       PortDirection.In, port_type)
        if self.debug:
            self.definition.add_debug_info(name, get_fn_ln())

    def output(self, name, width, is_signed=False, size=1,
               port_type=PortType.Data):
        self.definition.add_definition(name, width, size, is_signed,
                                       PortDirection.Out, port_type)
        if self.debug:
            self.definition.add_debug_info(name, get_fn_ln())

    def clock(self, name, is_input=True):
        if is_input:
            self.input(name, 1, port_type=PortType.Clock)
        else:
            self.output(name, 1, port_type=PortType.Clock)

    def reset(self, name, is_input=True):
        if is_input:
            self.input(name, 1, port_type=PortType.AsyncReset)
        else:
            self.output(name, 1, port_type=PortType.AsyncReset)

    def flip(self):
        bundle = PortBundle()
        bundle.definition = self.definition.flip()
        bundle.debug = self.debug
        bundle.definition.name = self.definition.name
        return bundle
